from typing import Callable, Optional


def _noop_udf_internal(
    fn: Optional[Callable] = None,
    **kwargs,
):
    if fn is not None:
        return fn
    else:
        return lambda fn2: fn2


udf = _noop_udf_internal



import sys
import time
import warnings
from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    Coroutine,
    Dict,
    Literal,
    Optional,
    Union,
    overload,
)

from fused._options import options as OPTIONS

if TYPE_CHECKING:
    import geopandas as gpd
    import pandas as pd
    import xarray as xr

from loguru import logger

from fused._load_udf import load
from fused._udf.decorators import _parse_ttl
from fused.models.api import UdfAccessToken, UdfJobStepConfig
from fused.models.api.udf_access_token import is_udf_token
from fused.models.udf import GeoPandasUdfV2
from fused.models.udf._eval_result import UdfEvaluationResult
from fused.types import UdfRuntimeError
from fused.warnings import FusedDeprecationWarning, FusedIgnoredWarning, FusedWarning

from .core import (
    run_file,
    run_file_async,
    run_shared_file,
    run_shared_file_async,
    run_shared_tile,
    run_shared_tile_async,
    run_tile,
    run_tile_async,
)
from .core._impl._realtime_ops_impl import default_run_engine

ResultType = Union["xr.Dataset", "pd.DataFrame", "gpd.GeoDataFrame"]

RUN_KWARGS = {
    "x",
    "y",
    "z",
    "sync",
    "engine",
    "type",
    "max_retry",
    "cache_max_age",
    "parameters",
    "_return_response",
}


@overload
def run(
    udf: Union[str, None, UdfJobStepConfig, GeoPandasUdfV2, UdfAccessToken] = None,
    *args,
    x: Optional[int] = None,
    y: Optional[int] = None,
    z: Optional[int] = None,
    sync: Literal[True] = True,
    engine: Optional[Literal["remote", "local"]] = None,
    type: Optional[Literal["tile", "file"]] = None,
    max_retry: int = 0,
    cache_max_age: Optional[str] = None,
    cache: bool = True,
    parameters: Optional[Dict[str, Any]] = None,
    _return_response: Literal[False] = False,
    _ignore_unknown_arguments: bool = False,
    **kw_parameters,
) -> ResultType: ...


@overload
def run(
    udf: Union[str, None, UdfJobStepConfig, GeoPandasUdfV2, UdfAccessToken] = None,
    *args,
    x: Optional[int] = None,
    y: Optional[int] = None,
    z: Optional[int] = None,
    sync: Literal[True] = True,
    engine: Optional[Literal["remote", "local"]] = None,
    type: Optional[Literal["tile", "file"]] = None,
    max_retry: int = 0,
    cache_max_age: Optional[str] = None,
    cache: bool = True,
    parameters: Optional[Dict[str, Any]] = None,
    _return_response: Literal[True] = True,
    _ignore_unknown_arguments: bool = False,
    **kw_parameters,
) -> UdfEvaluationResult: ...


@overload
def run(
    udf: Union[str, None, UdfJobStepConfig, GeoPandasUdfV2, UdfAccessToken] = None,
    *args,
    x: Optional[int] = None,
    y: Optional[int] = None,
    z: Optional[int] = None,
    sync: Literal[False] = False,
    engine: Optional[Literal["remote", "local"]] = None,
    type: Optional[Literal["tile", "file"]] = None,
    max_retry: int = 0,
    cache_max_age: Optional[str] = None,
    cache: bool = True,
    parameters: Optional[Dict[str, Any]] = None,
    _return_response: Literal[False] = False,
    _ignore_unknown_arguments: bool = False,
    **kw_parameters,
) -> Coroutine[ResultType, None, None]: ...


def run(
    udf: Union[str, None, UdfJobStepConfig, GeoPandasUdfV2, UdfAccessToken] = None,
    *args,
    x: Optional[int] = None,
    y: Optional[int] = None,
    z: Optional[int] = None,
    sync: bool = True,
    engine: Optional[Literal["remote", "local"]] = None,
    type: Optional[Literal["tile", "file"]] = None,
    max_retry: int = 0,
    cache_max_age: Optional[str] = None,
    cache: bool = True,
    parameters: Optional[Dict[str, Any]] = None,
    _return_response: Optional[bool] = False,
    _ignore_unknown_arguments: bool = False,
    **kw_parameters,
) -> Union[
    ResultType,
    Coroutine[ResultType, None, None],
    UdfEvaluationResult,
    Coroutine[UdfEvaluationResult, None, None],
]:
    """
    Executes a user-defined function (UDF) with various execution and input options.

    This function supports executing UDFs in different environments (local or remote),
    with different types of inputs (tile coordinates, geographical bounding boxes, etc.), and
    allows for both synchronous and asynchronous execution. It dynamically determines the execution
    path based on the provided parameters.

    Args:
        udf (str, GeoPandasUdfV2 or UdfJobStepConfig): the UDF to execute.
            The UDF can be specified in several ways:
            - A string representing a UDF name or UDF shared token.
            - A UDF object.
            - A UdfJobStepConfig object for detailed execution configuration.
        x, y, z: Tile coordinates for tile-based UDF execution.
        sync: If True, execute the UDF synchronously. If False, execute asynchronously.
        engine: The execution engine to use ('remote' or 'local').
        type: The type of UDF execution ('tile' or 'file').
        max_retry: The maximum number of retries to attempt if the UDF fails.
            By default does not retry.
        cache_max_age: The maximum age when returning a result from the cache.
            Supported units are seconds (s), minutes (m), hours (h), and days (d) (e.g. “48h”, “10s”, etc.).
            Default is `None` so a UDF run with `fused.run()` will follow `cache_max_age` defined in `@fused.udf()` unless this value is changed.
        cache: Set to False as a shortcut for `cache_max_age='0s'` to disable caching.
        parameters: Additional parameters to pass to the UDF.
        **kw_parameters: Additional parameters to pass to the UDF.

    Raises:
        ValueError: If the UDF is not specified or is specified in more than one way.
        TypeError: If the first parameter is not of an expected type.
        Warning: Various warnings are issued for ignored parameters based on the execution path chosen.

    Returns:
        The result of the UDF execution, which varies based on the UDF and execution path.

    Examples:
        Run a UDF saved in the Fused system:
        ```py
        fused.run("username@fused.io/my_udf_name")
        ```

        Run a UDF saved in GitHub:
        ```py
        loaded_udf = fused.load("https://github.com/fusedio/udfs/tree/main/public/Building_Tile_Example")
        fused.run(loaded_udf, bbox=bbox)
        ```

        Run a UDF saved in a local directory:
        ```py
        loaded_udf = fused.load("/Users/local/dir/Building_Tile_Example")
        fused.run(loaded_udf, bbox=bbox)
        ```

    Note:
        This function dynamically determines the execution path and parameters based on the inputs.
        It is designed to be flexible and support various UDF execution scenarios.
    """
    from fused._optional_deps import HAS_GEOPANDAS, HAS_MERCANTILE, HAS_SHAPELY

    job_step: Optional[UdfJobStepConfig] = None
    token: Optional[str] = None
    udf_email: Optional[str] = None
    udf_name: Optional[str] = None

    if "token" in kw_parameters:
        token = kw_parameters.pop("token")
        if udf is not None:
            warnings.warn(
                "token parameter is being ignored in favor of the first positional parameter.",
                FusedIgnoredWarning,
            )
        else:
            udf = token
        warnings.warn(
            "The 'token' keyword is deprecated. You can pass the token as the first "
            "argument instead (i.e. replace 'fused.run(token=<token>)' with "
            "'fused.run(<token>)')",
            FusedDeprecationWarning,
        )

    elif "udf_email" in kw_parameters and "udf_name" in kw_parameters:
        udf_email = kw_parameters.pop("udf_email")
        udf_name = kw_parameters.pop("udf_name")
        if udf is not None:
            warnings.warn(
                "udf_email parameter is being ignored in favor of the first positional parameter.",
                FusedIgnoredWarning,
            )
        else:
            udf = f"{udf_email}/{udf_name}"
            warnings.warn(
                "The 'udf_email' and 'udf_name' keywords are deprecated. You can pass "
                "the email and name as the first argument instead (i.e. replace "
                "'fused.run(udf_email=<email>, udf_name=<name>)' with "
                "'fused.run(\"<email>/<name>\")')",
                FusedDeprecationWarning,
            )

    elif args:
        if len(args) > 1 or not isinstance(args[0], str):
            raise TypeError(
                f"run() takes from 0 to 2 positional arguments but {1 + len(args)} were given"
            )
        udf_name = args[0]
        udf = f"{udf}/{udf_name}"
        warnings.warn(
            "The separate 'udf_email' and 'udf_name' arguments are deprecated. You can "
            "pass the email and name as the first argument instead (i.e. replace "
            "'fused.run(<email>, <name>)' with 'fused.run(\"<email>/<name>\")')",
            FusedDeprecationWarning,
        )

    if udf is None:
        raise ValueError("No UDF specified")

    if isinstance(udf, UdfJobStepConfig):
        job_step = udf
        udf = udf.udf
        udf_storage = "local_job_step"
    elif isinstance(udf, GeoPandasUdfV2):
        job_step = UdfJobStepConfig(udf=udf)
        udf_storage = "local_job_step"
    elif isinstance(udf, UdfAccessToken):
        token = udf.token
        udf = udf.token
        udf_storage = "token"
    elif isinstance(udf, str):
        if "/" in udf:
            udf_email, udf_name = udf.split("/", maxsplit=1)
            udf_storage = "saved"
        elif "@" in udf:
            udf_email = udf
            udf_storage = "saved"
        elif is_udf_token(udf):
            token = udf
            udf_storage = "token"
        else:
            # This will actually be the udf name, not the user's email
            udf_email = udf
            udf_storage = "saved"
    else:
        raise TypeError(
            "Could not detect UDF from first parameter. It should be a string, UdfJobStepConfig, or BaseUdf object."
        )

    if engine == "realtime":
        engine = "remote"
    elif engine == "batch":
        warnings.warn(
            "The 'batch' engine option is deprecated. Use fused.submit() instead.",
            FusedDeprecationWarning,
            stacklevel=2,
        )
    elif engine is None:
        if udf_storage in ["token", "saved"]:
            engine = "remote"
        else:
            engine = default_run_engine()
    elif engine not in ("local", "remote"):
        raise ValueError("Invalid engine specified. Must be 'local' or 'remote'.")

    # Loading a saved UDF
    if engine == "local" and udf_storage in ["token", "saved"]:
        try:
            loaded_udf = load(udf)
        except Exception as exc:
            raise ValueError(
                "Could not load UDF. Make sure the UDF is available locally when "
                'using `engine="local"`.\n'
                f'Error loading the UDF: "{exc}"'
            )
        job_step = UdfJobStepConfig(udf=loaded_udf)
        udf_storage = "local_job_step"
        # to avoid loading it later again
        udf = loaded_udf

    local_tile_bbox: Optional["gpd.GeoDataFrame"] = None
    xyz_ignored = False
    if x is not None and y is not None and z is not None:
        if HAS_MERCANTILE and HAS_GEOPANDAS and HAS_SHAPELY:
            import geopandas as gpd
            import mercantile
            import shapely

            tile_bounds = mercantile.bounds(x, y, z)
            local_tile_bbox = gpd.GeoDataFrame(
                {"x": [x], "y": [y], "z": [z]},
                geometry=[shapely.box(*tile_bounds)],
                crs=4326,
            )
        else:
            xyz_ignored = True

    if x is not None and y is not None and z is not None:
        if type != "tile":
            if type is None:
                # by default we still interpret x/y/z (if all passed) as a tile run
                type = "tile"
            else:
                kw_parameters["x"] = x
                kw_parameters["y"] = y
                kw_parameters["z"] = z
    else:
        if type is None:
            type = "file"
        elif type != "file":
            raise ValueError(
                "x, y, z not specified but type is 'tile', which is an invalid configuration. You must specify x, y, and z."
            )
        if x is not None:
            kw_parameters["x"] = x
        if y is not None:
            kw_parameters["y"] = y
        if z is not None:
            kw_parameters["z"] = z

    parameters = {
        **kw_parameters,
        **(parameters if parameters is not None else {}),
    }

    # Raise an error if passing a different argument than those expected from
    # the union of the UDF own parameters and the run() function parameters
    if parameters and not _ignore_unknown_arguments:
        if not isinstance(udf, GeoPandasUdfV2):
            try:
                udf = load(udf)
            except Exception:
                pass

        if isinstance(udf, GeoPandasUdfV2) and udf._parameter_list:
            udf_kwargs = set(udf._parameter_list)
            allowed_kwargs = udf_kwargs.union(
                RUN_KWARGS if type == "tile" else RUN_KWARGS - {"x", "y", "z"}
            )
            kw_params = set(parameters.keys())

            if not kw_params.issubset(allowed_kwargs):
                unexpected_args = kw_params - allowed_kwargs
                raise TypeError(
                    f"{udf.entrypoint}() got unexpected keyword argument '{unexpected_args.pop()}'"
                )

    cache_max_age = _parse_ttl(cache_max_age, cache)

    dispatch_params = ("sync" if sync else "async", type, udf_storage, engine)
    # Saved UDF
    if dispatch_params == ("sync", "tile", "saved", "remote"):
        fn = partial(
            run_tile,
            udf_email,
            udf_name,
            x=x,
            y=y,
            z=z,
            cache_max_age=cache_max_age,
            **parameters,
        )
    elif dispatch_params == ("async", "tile", "saved", "remote"):
        fn = partial(
            run_tile_async,
            udf_email,
            udf_name,
            x=x,
            y=y,
            z=z,
            cache_max_age=cache_max_age,
            **parameters,
        )
    elif dispatch_params == ("sync", "file", "saved", "remote"):
        fn = partial(
            run_file,
            udf_email,
            udf_name,
            cache_max_age=cache_max_age,
            **parameters,
        )
    elif dispatch_params == ("async", "file", "saved", "remote"):
        fn = partial(
            run_file_async,
            udf_email,
            udf_name,
            cache_max_age=cache_max_age,
            **parameters,
        )

    # shared UDF token
    elif dispatch_params == ("sync", "tile", "token", "remote"):
        fn = partial(
            run_shared_tile,
            token,
            x=x,
            y=y,
            z=z,
            cache_max_age=cache_max_age,
            **parameters,
        )
    elif dispatch_params == ("async", "tile", "token", "remote"):
        fn = partial(
            run_shared_tile_async,
            token,
            x=x,
            y=y,
            z=z,
            cache_max_age=cache_max_age,
            **parameters,
        )
    elif dispatch_params == ("sync", "file", "token", "remote"):
        fn = partial(run_shared_file, token, cache_max_age=cache_max_age, **parameters)
    elif dispatch_params == ("async", "file", "token", "remote"):
        fn = partial(
            run_shared_file_async, token, cache_max_age=cache_max_age, **parameters
        )

    # Local job step, which includes locally held code
    elif dispatch_params == ("sync", "tile", "local_job_step", "remote"):
        fn = lambda: job_step.run_tile(
            x=x, y=y, z=z, cache_max_age=cache_max_age, **parameters
        )
    elif dispatch_params == ("async", "tile", "local_job_step", "remote"):
        fn = lambda: job_step.run_tile_async(
            x=x, y=y, z=z, cache_max_age=cache_max_age, **parameters
        )
    elif dispatch_params == ("sync", "file", "local_job_step", "remote"):
        fn = lambda: job_step.run_file(cache_max_age=cache_max_age, **parameters)
    elif dispatch_params == ("async", "file", "local_job_step", "remote"):
        fn = lambda: job_step.run_file_async(**parameters)
    elif dispatch_params == ("sync", "tile", "local_job_step", "local"):
        fn = lambda: job_step.run_local(
            local_tile_bbox, **parameters, _return_response=_return_response
        )
    elif dispatch_params == ("async", "tile", "local_job_step", "local"):
        fn = lambda: job_step.run_local(local_tile_bbox, **parameters)
    elif dispatch_params == ("sync", "file", "local_job_step", "local"):
        fn = lambda: job_step.run_local(**parameters, _return_response=_return_response)
    elif dispatch_params == ("async", "file", "local_job_step", "local"):
        fn = lambda: job_step.run_local(**parameters)
    elif dispatch_params == ("sync", "tile", "local_job_step", "batch"):
        fn = None
    elif dispatch_params == ("sync", "file", "local_job_step", "batch"):
        fn = lambda: job_step.set_udf(job_step.udf, parameters=parameters).run_remote()
    else:
        fn = ...

    if xyz_ignored and engine == "local":
        # This warning doesn't matter on realtime because we will just put the x/y/z into the URL
        warnings.warn(
            FusedIgnoredWarning(
                "x, y, z arguments will be ignored because the following packages were not all found: mercantile shapely geopandas"
            ),
        )

    # Ellipsis is the sentinal value for not found in the dictionary at all
    if fn is Ellipsis:
        if udf_storage == "token" and engine != "remote":
            raise ValueError("UDF tokens can only be called on the 'remote' engine.")
        elif udf_storage == "saved" and engine != "remote":
            raise ValueError(
                "Saved UDFs can only be called on the 'remote' engine. To use another engine, load the UDF locally first."
            )
        else:
            raise ValueError(
                f"Could not determine how to call with settings: {dispatch_params}"
            )
    if fn is None:
        raise ValueError(f"Call type is not yet implemented: {dispatch_params}")
    if not sync and udf_storage not in ["saved", "token"]:
        raise ValueError(
            "Asynchronous execution is only supported for saved UDFs specifed by name or token."
        )

    n_retries = 0
    while n_retries <= max_retry:
        try:
            udf_eval_result = fn()
            break
        except Exception as exc:
            if n_retries >= max_retry:
                raise exc
            delay = OPTIONS.request_retry_base_delay * (2**n_retries)
            n_retries += 1
            warnings.warn(
                f"UDF execution failed, retrying in {delay} seconds (error: {exc})",
                FusedWarning,
            )

            time.sleep(delay)
            continue

    # Nested and remote UDF calls will return UdfEvaluationResult.
    # merge the stdout/stderr from fused.run() into running environment,
    # then return the UdfEvaluationResult object.
    if _return_response:
        return udf_eval_result
    if isinstance(udf_eval_result, UdfEvaluationResult):
        if udf_eval_result.stdout:
            sys.stdout.write(udf_eval_result.stdout)
        if udf_eval_result.stderr:
            sys.stderr.write(udf_eval_result.stderr)
        if udf_eval_result.is_cached:
            sys.stderr.write("Cached UDF result returned.\n")
            logger.info(f"Cache source: {udf_eval_result.cache_source.value}")
        if udf_eval_result.error_message is not None:
            raise UdfRuntimeError(
                udf_eval_result.error_message,
                child_exception_class=udf_eval_result.exception_class,
            )
        sys.stderr.flush()
        return udf_eval_result.data

    return udf_eval_result



import ast
import inspect
import re
import warnings
from pathlib import Path
from types import ModuleType
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

from fused._options import options as OPTIONS
from fused._request import session_with_retries
from fused._str_utils import is_url
from fused.models import Schema
from fused.models.udf import GeoPandasUdfV2, Header
from fused.models.udf._udf_registry import UdfRegistry
from fused.warnings import FusedDefaultWarning

DECORATOR_NAMES = "udf"


def _detect_function_name(src: str) -> Optional[str]:
    parsed = ast.parse(src)

    found_function_name: Optional[str] = None
    show_warning = False

    for node in ast.iter_child_nodes(parsed):
        if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            # Check if the function has decorators
            if node.decorator_list:
                # Find Fused decorator from the function
                for decorator in node.decorator_list:
                    new_found_function_name: Optional[str] = None
                    if isinstance(decorator, ast.Call):
                        if decorator.func.attr in DECORATOR_NAMES:
                            new_found_function_name = node.name
                    elif isinstance(decorator, ast.Attribute):
                        if decorator.attr in DECORATOR_NAMES:
                            # This is just in case there are no parens on the decorator.
                            new_found_function_name = node.name

                    if new_found_function_name:
                        if found_function_name:
                            show_warning = True
                        else:
                            found_function_name = new_found_function_name

    if show_warning:
        warnings.warn(
            FusedDefaultWarning(f"Loading the first UDF ({found_function_name})")
        )

    return found_function_name


def _detect_table_schema(src: str, function_name: str) -> Optional[str]:
    parsed = ast.parse(src)

    for node in ast.iter_child_nodes(parsed):
        if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            # Check if the function has decorators
            if node.name == function_name and node.decorator_list:
                # Find Fused decorator from the function
                for decorator in node.decorator_list:
                    if isinstance(decorator, ast.Call):
                        if decorator.func.attr in DECORATOR_NAMES:
                            for keyword in decorator.keywords:
                                if keyword.arg == "schema":
                                    if isinstance(keyword.value, ast.Str):
                                        return keyword.value.value
                                    elif isinstance(keyword.value, ast.Dict):
                                        if keyword.value.keys is None or not len(
                                            keyword.value.keys
                                        ):
                                            # Empty dictionary, special case
                                            return {}
                                    warnings.warn(
                                        FusedDefaultWarning(
                                            "A table schema was present but it could not be parsed. Use `eval_schema` to evaluate it."
                                        )
                                    )
                            return None
                    elif isinstance(decorator, ast.Attribute):
                        if decorator.attr in DECORATOR_NAMES:
                            # This is just in case there are no parens on the decorator.
                            return None

    return None


def _detect_parameters(src: str) -> Tuple[Dict[str, Any], List[str]]:
    def extract_arg_type(node):
        if isinstance(node, ast.Str):
            return node.s
        elif isinstance(node, ast.Tuple):
            return tuple(extract_arg_type(item) for item in node.elts)
        elif isinstance(node, ast.List):
            return [extract_arg_type(item) for item in node.elts]
        elif isinstance(node, ast.Num):
            return node.n
        elif isinstance(node, ast.Dict):
            return {k.s: extract_arg_type(v) for k, v in zip(node.keys, node.values)}

    try:
        parsed_ast = ast.parse(src)
    except SyntaxError:
        parsed_ast = ast.parse(repr(src))

    # Find and extract parameters
    params = []
    for node in ast.walk(parsed_ast):
        if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            for arg in node.args.args:
                params.append(arg.arg)
            break

    # Find and extract default values
    defaults = []
    for node in ast.walk(parsed_ast):
        if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            for default in node.args.defaults:
                defaults.append(extract_arg_type(default))
            break

    # Create a dictionary with parameters and their default values
    return dict(zip(params[-len(defaults) :], defaults)), params


def load_udf(
    udf_paths: Sequence[str],
    *,
    parameters: Optional[Dict[str, Any]] = None,
    content_type: Optional[str] = None,
    load_schema: bool = True,
    header_paths: Optional[Sequence[Header]] = None,
) -> UdfRegistry:
    """
    Load UDF(s) in a UdfRegistry object.

    Args:
        udf_paths: The paths to the UDF source code files or URLs.
            If provided as a list, it loads and registers multiple UDFs as a UdfRegistry.
        function: The name of the UDF function to load.
        parameters: A dictionary of parameters to be passed to the UDF.
        table_schema: The schema of the input data table.
        content_type: The content type of the UDF source, e.g., "file", "py", or "url".
        load_schema: Whether to automatically detect and load the table schema.
        header_paths: A sequence of headers for the UDF.

    Returns:
        UdfRegistry or UDF: Returns a UdfRegistry containing registered UDFs.

    Raises:
    - ValueError: If multiple UDFs with the same name are found in a list of UDF paths.
    - AssertionError: If an unsupported content type is provided.

    Examples:
        Load multiple UDFs from a list of files and register them in a UdfRegistry:

        ```py
        load_udf(udf_paths=["udf1.py", "udf2.py"], header_paths=["header.py"])
        load_udf("my_udf.py", function="my_function", content_type="file")
        ```
    """
    if not isinstance(udf_paths, list):
        udf = _load_udfs_from_file(
            udf_paths=udf_paths,
            parameters=parameters,
            content_type=content_type,
            load_schema=load_schema,
            header_paths=header_paths,
        )
        if isinstance(udf, UdfRegistry):
            return udf
        else:
            return UdfRegistry({udf.name: udf})
    else:
        udfs = {}
        for udf_path in udf_paths:
            udf = _load_udfs_from_file(
                udf_path,
                parameters=parameters,
                content_type=content_type,
                load_schema=load_schema,
                header_paths=header_paths,
            )

            # If the returned object is already a registry, just return it.
            if isinstance(udf, UdfRegistry):
                return udf
            if udf.name not in udfs:
                udfs[udf.name] = udf
            else:
                raise ValueError(
                    f"Multiple UDFs with the same name ({udf.name}) are not allowed."
                )
        return UdfRegistry(udfs)


def _load_udfs_from_file(
    udf_paths: str,
    function: Optional[str] = None,
    *,
    parameters: Optional[Dict[str, Any]] = None,
    table_schema: Union[Dict[str, Any], Schema, str, None] = None,
    content_type: Optional[str] = None,
    load_schema: bool = True,
    header_paths: Optional[Sequence[Header]] = None,
) -> Union[UdfRegistry, GeoPandasUdfV2]:
    """
    Load UDF(s) from a single source.

    Parameters:
    - udf_paths (str): The path to the UDF source code files or URLs.
    - function (str, optional): The name of the UDF function to load.
    - parameters (dict, optional): A dictionary of parameters to be passed to the UDF.
    - table_schema (dict, Schema, str, or None, optional): The schema of the input data table.
    - content_type (str, optional): The content type of the UDF source, e.g., "file", "py", or "url".
    - load_schema (bool, optional): Whether to automatically detect and load the table schema.
    - header_paths (Sequence[Header], optional): A sequence of headers for the UDF.

    Returns:
    - UdfRegistry or UDF: Returns an instantiated UDF object or a UdfRegistry if multiple UDFs
      were found.

    Raises:
    - ValueError: If multiple UDFs with the same name are found in a list of UDF paths.
    - AssertionError: If an unsupported content type is provided.
    """
    src = None
    _is_url = False

    if content_type == "url" or (content_type is None and is_url(udf_paths)):
        _is_url = True
    elif content_type is None or content_type == "file":
        udf_paths = Path(udf_paths)
    elif content_type == "py":
        src = udf_paths
    else:
        assert False, (
            f'Unknown content type: {content_type}. Should be one of "file", "py".'
        )

    if isinstance(udf_paths, Path):
        src = Path(udf_paths).read_text("utf-8")
    elif isinstance(udf_paths, ModuleType):
        if function:
            assert hasattr(udf_paths, function), "Could not find function on object."
        src = inspect.getsource(udf_paths)
    elif _is_url:
        with session_with_retries() as session:
            r = session.get(udf_paths, timeout=OPTIONS.request_timeout)
            r.raise_for_status()
            src = r.text
    else:
        pass
        # assert False, f"Unknown type to load from: {type(udf_paths)}"

    assert src, "No source provided"

    # If the source file contains multiple UDFs
    udfs = _find_udfs(src)
    if len(udfs.items()) > 1:
        # TODO: Update return type hint
        return load_udf(
            udf_paths=list(udfs.values()),
            header_paths=header_paths,
            load_schema=load_schema,
            content_type="py",
        )

    function = function if function else _detect_function_name(src)

    if table_schema:
        table_schema = table_schema
    elif load_schema:
        table_schema = _detect_table_schema(src, function)
    else:
        raise ValueError("No table schema provided and `load_schema` is False.")

    if isinstance(table_schema, str):
        table_schema = Schema.from_string(table_schema)
    elif table_schema is not None and not isinstance(table_schema, Schema):
        table_schema = Schema.model_validate(table_schema)

    # Set param headers on code string headers
    if header_paths:
        src = _replace_headers(src, header_paths)
    else:
        # If no `header_paths` but `src` contains headers, raise an error
        src_headers = _get_src_headers(src)
        if src_headers:
            raise ValueError(
                f"`load_udf` requires the `headers` parameter to be specified because the UDF source file specifies headers as `headers={src_headers}`."
            )

    _parameters, _parameter_list = _detect_parameters(src)
    new_udf = GeoPandasUdfV2(
        code=udfs[function].strip("\n"),  # UDF src without imports
        name=function,
        entrypoint=function,
        table_schema=table_schema,
        parameters=parameters or _parameters,
        headers=header_paths or [],
    )

    new_udf._parameter_list = _parameter_list
    return new_udf


def _find_udfs(src):
    fused_udf_functions = {}
    tree = ast.parse(src)
    for node in ast.walk(tree):
        if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            for decorator in node.decorator_list:
                if isinstance(decorator, ast.Call) and isinstance(
                    decorator.func, ast.Attribute
                ):
                    # Handle @fused.udf()
                    func = decorator.func
                elif isinstance(decorator, ast.Attribute):
                    # Handle @fused.udf
                    func = decorator
                else:
                    # If neither of the above, skip. This handles decorators internal to the udf.
                    continue
                if func.value.id == "fused" and func.attr == "udf":
                    # Get the code of the function, starting at the decorator.
                    start_lineno = decorator.lineno - 1
                    end_lineno = node.end_lineno
                    function_code = "\n".join(src.splitlines()[start_lineno:end_lineno])
                    fused_udf_functions[node.name] = function_code
    return fused_udf_functions


def _replace_headers(src: str, header_paths) -> str:
    def dict_from_file_paths(file_paths):
        file_dict = {}
        for file_path in file_paths:
            if isinstance(file_path, str):
                file_name = Path(file_path).stem
                file_dict[file_name] = file_path
            else:
                # TODO: Phase out
                file_dict[file_path.module_name + ".py"] = file_path
        return file_dict

    def replace_headers(src, header_paths):
        new_headers = str(header_paths)

        # Regex to match the headers parameter
        pattern = r"headers=\[[^\]]+\]"

        # Use regex to find and replace the headers value in the code
        return re.sub(pattern, f"headers={new_headers}", src)

    src_headers = _get_src_headers(src)
    if not src_headers:
        if header_paths:
            # warnings.warn(
            #     "Ignoring header specified in `load_udf` because UDF declaration does not specify header.",
            #     FusedDefaultWarning,
            # )
            pass
        else:
            return src

    # Assert that all headers in original UDF code are passed of the provided headers
    for src_header in set(dict_from_file_paths(src_headers).keys()):
        if src_header not in list(dict_from_file_paths(header_paths).keys()):
            ValueError(
                "The header {src_header} in UDF code do not match the headers provided."
            )
    return replace_headers(src, header_paths)


def _get_src_headers(src):
    # Regex to match the headers parameter
    pattern = r"headers=\s*\[([\s\S]*?)\]"

    # Use regex to find and extract the headers value
    match = re.search(pattern, src)

    # If a match is found, extract and split the headers into an array
    if match:
        return [header.strip("\"' ,\n\t") for header in match.group(1).split(",")]
    else:
        return []



import concurrent.futures
import json
import time
import warnings
from collections import defaultdict
from datetime import datetime, timedelta
from math import floor
from types import FunctionType
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Literal,
    Optional,
    Tuple,
    Union,
    overload,
)

from requests.exceptions import ReadTimeout

from fused._optional_deps import HAS_PANDAS, PD_DATAFRAME
from fused._options import options as OPTIONS
from fused._run import ResultType
from fused._run import run as fused_run
from fused._udf import udf as fused_udf
from fused.models.udf import GeoPandasUdfV2
from fused.types import UdfRuntimeError
from fused.warnings import FusedDeprecationWarning

if TYPE_CHECKING:
    import pandas as pd


Status = Literal["cancelled", "running", "timeout", "error", "success", "pending"]
STATUS_NOT_FINISHED = ("running", "pending")
STATUS_FAILED = ("error", "timeout")


def _coerce_hashable(df: "pd.DataFrame"):
    """Make the values in the DataFrame hashable.
    E.g. Python `list` cannot be hashed."""

    for index in df.index:
        for column in df.columns:
            existing_value = df.loc[index, column]

            if isinstance(existing_value, list):
                # list we know a way to make it a hashable type
                existing_value = tuple(existing_value)

            try:
                # Some other types like `dict` we don't have a strategy
                # for, so instead try coercing to json.
                hash(existing_value)
            except TypeError:
                existing_value = json.dumps(existing_value)

            df.loc[index, column] = existing_value

    return df


class Future:
    # or `Job`, `JobResult`, `Task`, ...

    def __init__(self, future, index, args):
        self._future: concurrent.futures.Future = future
        self._index = index
        self._args = args
        self._started_at = None
        self._ended_at = None

    def time(self) -> Optional[timedelta]:
        """How long this future took to complete, or None
        if the future is not complete."""
        if self._started_at is None or self._ended_at is None:
            return None
        return self._ended_at - self._started_at

    def done(self) -> bool:
        return self._future.done()

    def result(self):
        response = self._future.result()
        if isinstance(response, Exception):
            raise response
        if response.error_message is not None:
            raise self.exception()
        return response.data

    def exception(self) -> Optional[Exception]:
        try:
            response = self._future.result()
        except concurrent.futures.CancelledError as e:
            return e
        if isinstance(response, Exception):
            return response
        if response.error_message is not None:
            return UdfRuntimeError(
                f"[Run #{self._index} {self._args}] {response.error_message}",
                child_exception_class=response.exception_class,
            )
        return self._future.exception()

    def logs(self) -> str:
        response = self._future.result()
        if isinstance(response, ReadTimeout):
            return str(response)
        out = ""
        if response.stdout:
            out += "stdout\n------\n" + response.stdout
        if response.stderr:
            out += "\nstderr\n------\n" + response.stderr
        return out

    def status(self) -> Status:
        if self._future.cancelled():
            return "cancelled"
        elif self._future.running():
            return "running"
        elif self._future.done():
            exc = self.exception()
            if exc:
                if isinstance(
                    exc, ReadTimeout
                ) or "504 Server Error: Gateway Timeout" in str(exc):
                    return "timeout"
                return "error"
            return "success"  # or "finished"?
        else:
            return "pending"

    def __repr__(self) -> str:
        return f"<fused.Future [status: {('done - ' if self.done() else '') + self.status()}]>"


class JobPool:
    # or `Futures`, `PoolRunner`, etc

    """
    Pool of UDF runs. Don't use this class directly, use `fused.submit` instead.
    """

    def __init__(
        self,
        udf,
        arg_list,
        kwargs=None,
        engine="remote",
        max_workers=None,
        max_retry=2,
        before_run=None,
        wait_sleep=0.01,
        before_submit=0.01,
    ):
        self.udf = udf
        self.arg_list = arg_list
        self.n_jobs = len(self.arg_list)
        self._kwargs = kwargs or {}
        self._engine = engine
        self._pool = concurrent.futures.ThreadPoolExecutor(max_workers=max_workers)
        self._max_retry = max_retry
        self._before_run = before_run
        self._wait_sleep = wait_sleep
        self._before_submit = before_submit
        self._restarted_at = None
        self._started_at = None

    def _create_future(self, index, args) -> Future:
        future = Future(None, index, args)

        def _run(args):
            # currently we have to add a small delay between starting the UDF runs
            # to avoid overloading the server
            if self._before_run is not None:
                time.sleep(self._before_run)

            try:
                future._started_at = datetime.now()

                return fused_run(
                    self.udf,
                    engine=self._engine,
                    _return_response=True,
                    max_retry=self._max_retry,
                    **args,
                    **self._kwargs,
                )
            except Exception as exc:
                # ReadTime or HTTPError can happen on time-out or other server error
                return exc
            finally:
                future._ended_at = datetime.now()

        if self._before_submit:
            time.sleep(self._before_submit)

        future._future = self._pool.submit(_run, args)

        return future

    def _start_jobs(self):
        self._started_at = datetime.now()
        self._futures = [
            self._create_future(index, args) for index, args in enumerate(self.arg_list)
        ]

    def _get_status(self) -> List[bool]:
        return [f.done() for f in self._futures]

    def _status_counts(self) -> Dict[Status, int]:
        counts = defaultdict(int)
        for f in self._futures:
            counts[f.status()] += 1

        # TODO: always sort counts keys by status order
        return counts

    def _status_message(self) -> str:
        counts = self._status_counts()

        message_parts = []
        for status in ("running", "timeout", "error", "success", "pending"):
            c = counts[status]
            if c:
                message_parts.append(f"{c} {status}")

        if len(message_parts) == 0:
            return "empty"

        return ", ".join(message_parts)

    def _get_progress(self) -> Tuple[int, str]:
        # floor as to not show 100 if not every one is actually done
        percentage = floor(sum(self._get_status()) / self.n_jobs * 100)
        return percentage, f"{sum(self._get_status())}/{self.n_jobs}"

    def retry(self):
        """Rerun any tasks in error or timeout states. Tasks are rerun in the same pool."""
        old_futures = self._futures

        def _create_new_future(index, args):
            if old_futures[index].status() in ("error", "timeout"):
                return self._create_future(index, args)
            return old_futures[index]

        self._restarted_at = datetime.now()
        self._futures = [
            _create_new_future(index, args) for index, args in enumerate(self.arg_list)
        ]

    def cancel(self, wait: bool = True):
        """Cancel any pending (not running) tasks.

        Note it will not be possible to retry on the same JobPool later."""
        self._pool.shutdown(wait=wait, cancel_futures=True)

    def total_time(self, since_retry: bool = False) -> timedelta:
        """Returns how long the entire job took.

        If only partial results are available, returns based on the last task to have been completed.
        """
        started_at = (
            self._restarted_at
            if since_retry and self._restarted_at is not None
            else self._started_at
        )
        if started_at is None:
            raise ValueError("JobPool has not been started")
        all_result_at = [f._ended_at for f in self._futures if f._ended_at is not None]
        if not len(all_result_at):
            raise ValueError("No result is available yet")
        last_result_at = max(all_result_at)
        return last_result_at - self._started_at

    def times(self) -> list[Optional[timedelta]]:
        """Time taken for each task.

        Incomplete tasks will be reported as None."""
        return [f.time() for f in self._futures]

    def done(self) -> bool:
        """True if all tasks have finished, regardless of success or failure."""
        return all(fut.done() for fut in self._futures)

    def all_succeeded(self) -> bool:
        """True if all tasks finished with success"""
        return all(f.status() == "success" for f in self._futures)

    def any_failed(self) -> bool:
        """True if any task finished with an error"""
        return any(f.status() in ("error", "timeout") for f in self._futures)

    def any_succeeded(self) -> bool:
        """True if any task finished with success"""
        return any(f.status() == "success" for f in self._futures)

    def arg_df(self):
        """The arguments passed to runs as a DataFrame"""
        if not HAS_PANDAS:
            raise ImportError("pandas is required to use this method")

        import pandas as pd

        return pd.DataFrame(self.arg_list)

    def status(self):
        """Return a Series indexed by status of task counts"""
        return self.df().value_counts("status")

    def wait(self):
        """Wait until all jobs are finished

        Use fused.options.show.enable_tqdm to enable/disable tqdm.
        Use pool._wait_sleep to set if sleep should occur while waiting.
        """

        def _noop(*args, **kwargs):
            pass

        after = _noop
        update = _noop
        if OPTIONS.show.enable_tqdm:
            from tqdm.auto import tqdm

            t = tqdm(total=self.n_jobs, bar_format="{l_bar}{bar} | {n_fmt}/{total_fmt}")

            def _tqdm_after():
                t.update(self.n_jobs - t.n)
                t.close()

            after = _tqdm_after
            update = lambda n_done: t.update(n_done - t.n)

        while not self.done():
            n_done = sum(self._get_status())
            update(n_done)
            if self._wait_sleep is not None:
                time.sleep(self._wait_sleep)

        after()

    def tail(self, stop_on_exception=False):
        """Wait until all jobs are finished, printing statuses as they become available.

        This is useful for interactively watching for the state of the pool.

        Use pool._wait_sleep to set if sleep should occur while waiting.
        """
        seen = set(self.results_now(return_exceptions=not stop_on_exception).keys())
        if len(seen):
            print(self._status_message())

            if len(seen) == len(self._futures):
                # Nothing left to tail
                return

        while len(seen) != len(self._futures):
            done = self.results_now(return_exceptions=not stop_on_exception)
            for new_key in done.keys():
                if new_key not in seen:
                    seen.add(new_key)
                    # TODO: Logger support, rather than print statement?
                    print(
                        f"[{datetime.now()}] Run #{new_key} {self._futures[new_key]._args} {self._futures[new_key].status()} ({self._status_message()})"
                    )

            if self._wait_sleep is not None:
                time.sleep(self._wait_sleep)

        print(f"End of tail\n{self._status_message()}")

    def results(self, return_exceptions=False) -> List[Any]:
        """Retrieve all results of the job.

        Results are ordered by the order of the args list."""
        results = []
        for fut in self._futures:
            try:
                results.append(fut.result())
            except Exception:
                if return_exceptions:
                    results.append(fut.exception())
                else:
                    raise
        return results

    def results_now(self, return_exceptions=False) -> Dict[int, Any]:
        """Retrieve the results that are currently done.

        Results are indexed by position in the args list."""
        results = {}
        for index, fut in enumerate(self._futures):
            if fut.done():
                try:
                    results[index] = fut.result()
                except Exception:
                    if return_exceptions:
                        results[index] = fut.exception()
                    else:
                        raise
        return results

    def df(
        self,
        status_column: Optional[str] = "status",
        result_column: Optional[str] = "result",
        time_column: Optional[str] = "time",
        logs_column: Optional[str] = "logs",
        exception_column: Optional[str] = None,
        include_exceptions: bool = True,
    ):
        """
        Get a DataFrame of results as they are currently.
        The DataFrame will have columns for each argument passed, and columns for:
        `status`, `result`, `time`, `logs` and optionally `exception`.
        """
        if not HAS_PANDAS:
            raise ImportError("pandas is required to use this method")

        import pandas as pd

        results_df = pd.DataFrame(self.arg_list)
        status_list = [f.status() for f in self._futures]
        if status_column:
            results_df[status_column] = status_list

        if time_column:
            time_list = [None] * len(self._futures)
            for i, fut in enumerate(self._futures):
                t = fut.time()
                if t is not None:
                    time_list[i] = t.total_seconds()

            results_df[time_column] = pd.Series(time_list, dtype=float)

        if result_column:
            result_list = [None] * len(self._futures)
            for i, fut in enumerate(self._futures):
                if status_list[i] not in STATUS_NOT_FINISHED:
                    try:
                        res = fut.result()
                    except Exception:
                        res = fut.exception() if include_exceptions else None

                    result_list[i] = res

            results_df[result_column] = result_list

        if logs_column:
            logs_list = [None] * len(self._futures)
            for i, fut in enumerate(self._futures):
                if status_list[i] not in STATUS_NOT_FINISHED:
                    logs_list[i] = fut.logs()

            results_df[logs_column] = logs_list

        if exception_column:
            results_df[exception_column] = [fut.exception() for fut in self._futures]

        return results_df

    def get_status_df(self):
        warnings.warn(
            "the 'get_status_df()' method is deprecated, use '.df()' instead.",
            FusedDeprecationWarning,
        )
        return self.df()

    def get_results_df(self, ignore_exceptions=False):
        warnings.warn(
            "the 'get_results_df()' method is deprecated, use '.df()' instead.",
            FusedDeprecationWarning,
        )
        self.wait()
        return self.df(include_exceptions=not ignore_exceptions)

    def errors(self) -> Dict[int, Exception]:
        """Retrieve the results that are currently done and are errors.

        Results are indexed by position in the args list."""
        errors = {}
        for index, fut in enumerate(self._futures):
            if fut.status() in STATUS_FAILED:
                errors[index] = fut.exception()
        return errors

    def first_error(self) -> Optional[Exception]:
        """Retrieve the first (by order of arguments) error result, or None."""
        for fut in self._futures:
            if fut.status() in STATUS_FAILED:
                return fut.exception()

        return None

    def logs(self) -> list[str]:
        """Logs for each task.

        Incomplete tasks will be reported as None."""
        return [
            (f.logs() if f.status() not in STATUS_NOT_FINISHED else None)
            for f in self._futures
        ]

    def first_log(self) -> Optional[str]:
        """Retrieve the first (by order of arguments) logs, or None."""
        for f in self._futures:
            if f.status() not in STATUS_NOT_FINISHED:
                return f.logs()

    def success(self) -> Dict[int, Any]:
        """Retrieve the results that are currently done and are successful.

        Results are indexed by position in the args list."""
        success = {}
        for index, fut in enumerate(self._futures):
            if fut.status() in ("success",):
                success[index] = fut.result()
        return success

    def pending(self) -> Dict[int, Any]:
        """Retrieve the arguments that are currently pending and not yet submitted."""
        pending = {}
        for index, fut in enumerate(self._futures):
            if fut.status() in ("pending",):
                pending[index] = fut._args
        return pending

    def running(self) -> Dict[int, Any]:
        """Retrieve the results that are currently running."""
        running = {}
        for index, fut in enumerate(self._futures):
            if fut.status() in ("running",):
                running[index] = fut._args
        return running

    def cancelled(self) -> Dict[int, Any]:
        """Retrieve the arguments that were cancelled and not run."""
        cancelled = {}
        for index, fut in enumerate(self._futures):
            if fut.status() in ("cancelled",):
                cancelled[index] = fut._args
        return cancelled

    def collect(self, ignore_exceptions=False, flatten=True):
        """Collect all results into a DataFrame"""
        if not HAS_PANDAS:
            raise ImportError("pandas is required to use this method")

        import pandas as pd

        results = self.results(return_exceptions=ignore_exceptions)
        mask = [not isinstance(r, Exception) for r in results]
        results = dict(
            res
            for res in zip(range(self.n_jobs), results)
            if not isinstance(res[1], Exception)
        )

        results_pandas = all(isinstance(res, pd.DataFrame) for res in results.values())
        if results_pandas and len(results) and flatten:
            # TODO: flatten can un-occur depending on results, this seems wrong
            results_df = pd.concat(results)
        else:
            results_df = pd.DataFrame({"result": pd.Series(results)})
            results_df.index = pd.MultiIndex.from_product([results_df.index])

        args_df = pd.DataFrame(self.arg_list)
        args_df = args_df[mask]

        _coerce_hashable(args_df)

        # combine concatenated results with arguments as prepended index levels
        assert len(results_df.index.levels[0]) == len(args_df)
        args_index = pd.MultiIndex.from_frame(args_df)
        args_codes = [c.take(results_df.index.codes[0]) for c in args_index.codes]
        new_idx = pd.MultiIndex(
            levels=list(args_index.levels) + results_df.index.levels[1:],
            codes=args_codes + results_df.index.codes[1:],
            names=list(args_df.columns) + results_df.index.names[1:],
        )
        if new_idx.nlevels == 1:
            new_idx = new_idx.get_level_values(0)
        results_df.index = new_idx
        return results_df

    def __getitem__(self, idx: int) -> Future:
        return self._futures[idx]

    def __len__(self) -> int:
        return self.n_jobs

    def __repr__(self) -> str:
        return f"<JobPool with {self.n_jobs} jobs [{self._status_message()}]>"

    def _repr_html_(self) -> str:
        # TODO we could provide a more informative repr in notebooks (e.g. showing
        # a table of the individual jobs and their status?)
        counts = self._status_counts()
        status_table_data = [
            f"<tr><td>{status}</td><td>{count}</td></tr>"
            for status, count in counts.items()
        ]
        return f"<table><tr><th>Status</th><th>Count</th></tr>{''.join(status_table_data)}</table>"


def _validate_arg_list(arg_list, udf):
    if HAS_PANDAS and isinstance(arg_list, PD_DATAFRAME):
        return arg_list.to_dict(orient="records")

    if not len(arg_list):
        raise ValueError("arg_list must be a non-empty list")
    if not isinstance(arg_list[0], dict):
        if not isinstance(udf, GeoPandasUdfV2):
            raise ValueError(
                "arg_list must be a list of dictionaries. A simple list to pass "
                "as first positional argument is only supported for UDF objects."
            )
        if udf._parameter_list is None:
            # TODO: Run _parameter_list detection here
            raise NotImplementedError(
                "arg_list must be a list of dictionaries. Could not detect the first "
                "positional argument name."
            )
        if len(udf._parameter_list) == 0:
            raise ValueError(
                "arg_list must be a list of dictionaries. UDF does not accept any arguments."
            )
        name = udf._parameter_list[0]
        arg_list = [{name: arg} for arg in arg_list]

    return arg_list


@overload
def submit(
    udf,
    arg_list,
    /,
    *,
    engine: Optional[Literal["remote", "local"]] = "remote",
    max_workers: Optional[int] = None,
    max_retry: int = 2,
    debug_mode: Literal[False] = False,
    collect: Literal[False] = False,
    **kwargs,
) -> JobPool: ...


@overload
def submit(
    udf,
    arg_list,
    /,
    *,
    engine: Optional[Literal["remote", "local"]] = "remote",
    max_workers: Optional[int] = None,
    max_retry: int = 2,
    debug_mode: Literal[True] = True,
    collect: bool = True,
    **kwargs,
) -> ResultType: ...


@overload
def submit(
    udf,
    arg_list,
    /,
    *,
    engine: Optional[Literal["remote", "local"]] = "remote",
    max_workers: Optional[int] = None,
    max_retry: int = 2,
    debug_mode: Literal[False] = False,
    collect: Literal[True] = True,
    **kwargs,
) -> "pd.DataFrame": ...


def submit(
    udf,
    arg_list,
    /,
    *,
    engine: Optional[Literal["remote", "local"]] = "remote",
    max_workers: Optional[int] = None,
    max_retry: int = 2,
    debug_mode: bool = False,
    collect: bool = True,
    cache_max_age: Optional[str] = None,
    cache: bool = True,
    ignore_exceptions: bool = False,
    flatten: bool = True,
    _before_run: Optional[float] = None,
    _before_submit: Optional[float] = 0.01,
    **kwargs,
) -> Union[JobPool, ResultType, "pd.DataFrame"]:
    """
    Executes a user-defined function (UDF) multiple times for a list of input
    parameters, and return immediately a "lazy" JobPool object allowing
    to inspect the jobs and wait on the results.

    See `fused.run` for more details on the UDF execution.

    Args:
        udf: the UDF to execute.
            See `fused.run` for more details on how tos specify the UDF.
        arg_list: a list of input parameters for the UDF. Can be specified as:
            - a list of values for parametrizing over a single parameter, i.e.
              the first parameter of the UDF
            - a list of dictionaries for parametrizing over multiple parameters
            - A DataFrame for parametrizing over multiple parameters where each
              row is a set of parameters

        engine: The execution engine to use. Defaults to 'remote'.
        max_workers: The maximum number of workers to use. Defaults to 32.
        max_retry: The maximum number of retries for failed jobs. Defaults to 2.
        debug_mode: If True, executes only the first item in arg_list directly using
            `fused.run()`, useful for debugging UDF execution. Default is False.
        collect: If True, waits for all jobs to complete and returns the collected DataFrame
            containing the results. If False, returns a JobPool object, which is non-blocking
            and allows you to inspect the individual results and logs.
            Default is True.
        cache_max_age: The maximum age when returning a result from the cache.
            Supported units are seconds (s), minutes (m), hours (h), and days (d)
            (e.g. “48h”, “10s”, etc.).
            Default is `None` so a UDF run with `fused.run()` will follow
            `cache_max_age` defined in `@fused.udf()` unless this value is changed.
        cache: Set to False as a shortcut for `cache_max_age='0s'` to disable caching.
        ignore_exceptions: Set to True to ignore exceptions when collecting results.
            Runs that result in exceptions will be silently ignored. Defaults to False.
        flatten: Set to True to receive a DataFrame of results, without nesting of a
            `results` column, when collecting results. When False, results will be nested
            in a `results` column. If the UDF does not return a DataFrame (e.g. a string
            instead,) results will be nested in a `results` column regardless of this setting.
            Defaults to True.
        **kwargs: Additional (constant) keyword arguments to pass to the UDF.

    Returns:
        JobPool

    Examples:
        Run a UDF multiple times for the values 0 to 9 passed to as the first
        positional argument of the UDF:
        ```py
        df = fused.submit("username@fused.io/my_udf_name", range(10))
        ```

        Being explicit about the parameter name:
        ```py
        df = fused.submit(udf, [dict(n=i) for i in range(10)])
        ```

        Get the pool of ongoing tasks:
        ```py
        pool = fused.submit(udf, [dict(n=i) for i in range(10)], collect=False)
        ```

    """
    if isinstance(udf, FunctionType):
        if udf.__name__ == "<lambda>":
            # This will not work correctly in fused.udf. If we find a way to parse just the AST
            # of the lambda (without any surrounding function call, assignment, etc.) then we can
            # support lambda here
            raise TypeError(
                """Lambda expressions cannot be passed into fused.submit(). Create a function with @fused.udf instead:
@fused.udf
def udf(x):
    return x

fused.submit(udf, arg_list)
"""
            )
        # TODO: Move this logic to fused.run?
        udf = fused_udf(udf)

    arg_list = _validate_arg_list(arg_list, udf)

    if cache_max_age is not None:
        kwargs["cache_max_age"] = cache_max_age
    if not cache:
        kwargs["cache"] = cache

    if debug_mode:
        return fused_run(udf, engine=engine, **arg_list[0], **kwargs)

    max_workers = min(max_workers or 32, 1024)

    job_pool = JobPool(
        udf,
        arg_list,
        kwargs,
        engine=engine,
        max_workers=max_workers,
        max_retry=max_retry,
        before_run=_before_run,
        before_submit=_before_submit,
    )
    job_pool._start_jobs()

    if collect:
        job_pool.wait()
        return job_pool.collect(ignore_exceptions=ignore_exceptions, flatten=flatten)
    else:
        return job_pool



from __future__ import annotations

import inspect
import json
import re
import shlex
import warnings
from enum import Enum
from pathlib import Path
from typing import IO, TYPE_CHECKING, Any, Literal, Union

if TYPE_CHECKING:
    import geopandas as gpd
    import pandas as pd

import numpy as np
import pyarrow as pa
import pyarrow.parquet as pq
from pydantic import BaseModel, StrictInt, field_validator
from pydantic import Field as PydanticField

from fused._optional_deps import GPD_GEODATAFRAME, HAS_GEOPANDAS
from fused.warnings import FusedDefaultWarning, FusedReservedWarning, FusedWarning

INDEX_COLUMN_NAME = "fused_index"
GEOARROW_NAME = "geoarrow.wkb"


class BaseDataType:
    """A base class for Fused DataType objects"""

    pass


class GeoArrowWKB(pa.ExtensionType):
    """A PyArrow extension type to store the "geoarrow.wkb" name on a binary column."""

    # This is derived from docs here:
    # https://arrow.apache.org/docs/python/generated/pyarrow.ExtensionType.html
    def __init__(self):
        pa.ExtensionType.__init__(self, pa.binary(), GEOARROW_NAME)

    def __arrow_ext_serialize__(self):
        # since we don't have a parameterized type, we don't need extra
        # metadata to be deserialized
        return b""

    @classmethod
    def __arrow_ext_deserialize__(self, storage_type, serialized):
        # return an instance of this subclass given the serialized
        # metadata.
        return GeoArrowWKB()


try:
    pa.register_extension_type(GeoArrowWKB())
except pa.ArrowKeyError:
    pass


class PrimitiveDataType(BaseDataType, str, Enum):
    """An enumeration of data types that can be used in columns."""

    Null = "null"
    Bool = "Bool"
    Int8 = "Int8"
    Int16 = "Int16"
    Int32 = "Int32"
    Int64 = "Int64"
    Uint8 = "Uint8"
    Uint16 = "Uint16"
    Uint32 = "Uint32"
    Uint64 = "Uint64"
    Float16 = "Float16"
    Float32 = "Float32"
    Float64 = "Float64"
    Date32 = "Date32"
    Date64 = "Date64"
    String = "String"
    LargeString = "LargeString"
    Binary = "Binary"
    LargeBinary = "LargeBinary"
    Geometry = "Geometry"
    """WKB-encoded geometry as a Binary DataType"""

    def __str__(self) -> str:
        return self.value

    def to_arrow(self) -> pa.DataType:
        """Convert this type to a [`pyarrow.DataType`][pyarrow.DataType]."""

        if self == PrimitiveDataType.Null:
            return pa.null()
        if self == PrimitiveDataType.Bool:
            return pa.bool_()
        if self == PrimitiveDataType.Int8:
            return pa.int8()
        if self == PrimitiveDataType.Int16:
            return pa.int16()
        if self == PrimitiveDataType.Int32:
            return pa.int32()
        if self == PrimitiveDataType.Int64:
            return pa.int64()
        if self == PrimitiveDataType.Uint8:
            return pa.uint8()
        if self == PrimitiveDataType.Uint16:
            return pa.uint16()
        if self == PrimitiveDataType.Uint32:
            return pa.uint32()
        if self == PrimitiveDataType.Uint64:
            return pa.uint64()
        if self == PrimitiveDataType.Float16:
            return pa.float16()
        if self == PrimitiveDataType.Float32:
            return pa.float32()
        if self == PrimitiveDataType.Float64:
            return pa.float64()
        if self == PrimitiveDataType.Date32:
            return pa.date32()
        if self == PrimitiveDataType.Date64:
            return pa.date64()
        if self == PrimitiveDataType.String:
            return pa.string()
        if self == PrimitiveDataType.LargeString:
            return pa.large_string()
        if self == PrimitiveDataType.Binary:
            return pa.binary()
        if self == PrimitiveDataType.LargeBinary:
            return pa.large_binary()
        if self == PrimitiveDataType.Geometry:
            return GeoArrowWKB()

        raise ValueError(f"unexpected type {self=}")

    @classmethod
    def from_arrow(cls, dtype: pa.DataType) -> PrimitiveDataType:
        """Construct this from a [`pyarrow.DataType`][pyarrow.DataType]."""
        if isinstance(dtype, pa.ExtensionType):
            if dtype.extension_name == GEOARROW_NAME:
                return cls.Geometry

            raise ValueError(f"Unknown extension type {dtype.extension_name}")

        if pa.types.is_null(dtype):
            return cls.Null
        if pa.types.is_boolean(dtype):
            return cls.Bool
        if pa.types.is_int8(dtype):
            return cls.Int8
        if pa.types.is_int16(dtype):
            return cls.Int16
        if pa.types.is_int32(dtype):
            return cls.Int32
        if pa.types.is_int64(dtype):
            return cls.Int64
        if pa.types.is_uint8(dtype):
            return cls.Uint8
        if pa.types.is_uint16(dtype):
            return cls.Uint16
        if pa.types.is_uint32(dtype):
            return cls.Uint32
        if pa.types.is_uint64(dtype):
            return cls.Uint64
        if pa.types.is_float16(dtype):
            return cls.Float16
        if pa.types.is_float32(dtype):
            return cls.Float32
        if pa.types.is_float64(dtype):
            return cls.Float64
        if pa.types.is_date32(dtype):
            return cls.Date32
        if pa.types.is_date64(dtype):
            return cls.Date64
        if pa.types.is_string(dtype):
            return cls.String
        if pa.types.is_large_string(dtype):
            return cls.LargeString
        if pa.types.is_binary(dtype):
            return cls.Binary
        if pa.types.is_large_binary(dtype):
            return cls.LargeBinary

        assert False, f"unimplemented type {dtype}"

    @classmethod
    def from_numpy_type_class(cls, dtype: type[np.generic]) -> PrimitiveDataType:
        """Construct this from a [`numpy.generic`][numpy.generic] instance."""

        assert issubclass(dtype, np.generic), (
            f"Expected dtype to be a subclass of np.generic, got {dtype}"
        )

        return cls.from_arrow(pa.from_numpy_dtype(dtype))

    @classmethod
    def from_numpy_type(cls, dtype: np.generic) -> PrimitiveDataType:
        """Construct this from a [`numpy.generic`][numpy.generic] instance."""

        assert isinstance(dtype, np.generic)
        return cls.from_arrow(pa.from_numpy_dtype(dtype))

    @classmethod
    def from_numpy_dtype(cls, dtype: np.dtype) -> PrimitiveDataType:
        """Construct this from a [`numpy.dtype`][numpy.dtype] instance."""

        assert isinstance(dtype, np.dtype)
        return cls.from_numpy_type_class(dtype.type)

    @classmethod
    def from_pandas_type(
        cls, dtype: pd.api.extensions.ExtensionDtype
    ) -> PrimitiveDataType:
        """Construct this from a [pandas dtype][pandas.api.extensions.ExtensionDtype] instance."""
        import pandas as pd

        pandas_df = pd.DataFrame({"col": pd.Series(dtype=dtype)})
        pyarrow_schema = pa.Schema.from_pandas(pandas_df, preserve_index=False)
        return cls.from_arrow(pyarrow_schema.field("col").type)

    @classmethod
    def from_python_type(cls, dtype: type[object]) -> PrimitiveDataType:
        if dtype is float:
            return PrimitiveDataType.Float64

        if dtype is int:
            return PrimitiveDataType.Int64

        if dtype is str:
            return PrimitiveDataType.String

        if dtype is bool:
            return PrimitiveDataType.Bool

        raise TypeError(f"Expected {dtype} to be a builtin python type")


class TimestampType(BaseDataType, BaseModel):
    type: Literal["Timestamp"] = "Timestamp"
    unit: Literal["s", "ms", "us", "ns"]
    tz: str | None = None

    def __str__(self) -> str:
        if self.tz is not None:
            return f"{self.type}[{self.unit}, {self.tz}]"

        return f"{self.type}[{self.unit}]"

    def to_arrow(self) -> pa.TimestampType:
        return pa.timestamp(unit=self.unit, tz=self.tz)

    @classmethod
    def from_arrow(cls, dtype: pa.TimestampType) -> TimestampType:
        return TimestampType(unit=dtype.unit, tz=dtype.tz)

    @classmethod
    def from_string(cls, s: str) -> TimestampType:
        """Parse from a string type description

        Examples:
            ```py
            s = "Timestamp[ms]"
            TimestampType.from_string(s)
            ```

            ```py
            s = "Timestamp[ms, America/New_York]"
            TimestampType.from_string(s)
            ```

        """
        start = "Timestamp["
        end = "]"
        assert s.startswith(start)
        assert s.endswith(end)
        inner_s = s[len(start) : -len(end)]
        split = inner_s.split(",", maxsplit=1)
        if len(split) == 1:
            unit = split[0]
            return TimestampType(unit=unit.strip())  # type: ignore

        else:
            assert len(split) == 2
            unit, tz = split
            return TimestampType(unit=unit.strip(), tz=tz.strip())  # type: ignore


class ListType(BaseDataType, BaseModel):
    type: Literal["List"] = "List"
    element_type: DataType

    def __str__(self) -> str:
        return f"{self.type}[{self.element_type}]"

    def to_arrow(self) -> pa.ListType:
        value_type = self.element_type.to_arrow()
        return pa.list_(value_type)

    @classmethod
    def from_arrow(cls, dtype: pa.ListType) -> ListType:
        return ListType(element_type=data_type_from_arrow(dtype.value_type))

    @classmethod
    def from_string(cls, s: str) -> ListType:
        """Parse from a string type description

        Examples:
            ```py
            s = "List[Float32]"
            ListType.from_string(s)
            ```
        """
        start = "List["
        end = "]"
        assert s.startswith(start)
        assert s.endswith(end)
        inner_s = s[len(start) : -len(end)]
        return ListType(element_type=data_type_from_string(inner_s))


class LargeListType(BaseDataType, BaseModel):
    type: Literal["LargeList"] = "LargeList"
    element_type: DataType

    def __str__(self) -> str:
        return f"{self.type}[{self.element_type}]"

    def to_arrow(self) -> pa.LargeListType:
        value_type = self.element_type.to_arrow()
        return pa.large_list(value_type)

    @classmethod
    def from_arrow(cls, dtype: pa.LargeListType) -> LargeListType:
        return LargeListType(element_type=data_type_from_arrow(dtype.value_type))

    @classmethod
    def from_string(cls, s: str) -> LargeListType:
        """Parse from a string type description

        Examples:
            ```py
            s = "LargeList[Uint8]"
            LargeListType.from_string(s)
            ```
        """
        start = "LargeList["
        end = "]"
        assert s.startswith(start)
        assert s.endswith(end)
        inner_s = s[len(start) : -len(end)]
        return LargeListType(element_type=data_type_from_string(inner_s))


class FixedSizeListType(BaseDataType, BaseModel):
    type: Literal["FixedSizeList"] = "FixedSizeList"
    element_type: DataType
    inner_size: StrictInt

    def __str__(self) -> str:
        return f"{self.type}[{self.element_type}, {self.inner_size}]"

    def to_arrow(self) -> pa.FixedSizeListType:
        value_type = self.element_type.to_arrow()
        return pa.list_(value_type, self.inner_size)

    @classmethod
    def from_arrow(cls, dtype: pa.FixedSizeListType) -> FixedSizeListType:
        return FixedSizeListType(
            element_type=data_type_from_arrow(dtype.value_type),
            inner_size=dtype.list_size,
        )

    @classmethod
    def from_string(cls, s: str) -> FixedSizeListType:
        """Parse from a string type description

        There should be two items in the bracket. The first should be the inner type.
        The second should be the number of inner elements of the fixed size list.

        Examples:
            ```py
            s = "FixedSizeList[Uint8, 4]"
            FixedSizeListType.from_string(s)
            ```
        """
        start = "FixedSizeList["
        end = "]"
        assert s.startswith(start)
        assert s.endswith(end)

        regex = r",\s*(\d+)\]$"
        match = re.search(regex, s)
        assert match
        inner_size = int(match.group(1))
        inner_str = s[len(start) : match.span(0)[0]]
        return FixedSizeListType(
            element_type=data_type_from_string(inner_str), inner_size=inner_size
        )


class FixedSizeBinaryType(BaseDataType, BaseModel):
    type: Literal["FixedSizeBinary"] = "FixedSizeBinary"
    byte_width: StrictInt

    def __str__(self) -> str:
        return f"{self.type}[{self.byte_width}]"

    def to_arrow(self) -> pa.FixedSizeBinaryType:
        return pa.binary(self.byte_width)

    @classmethod
    def from_arrow(cls, dtype: pa.FixedSizeBinaryType) -> FixedSizeBinaryType:
        return FixedSizeBinaryType(byte_width=dtype.byte_width)

    @classmethod
    def from_string(cls, s: str) -> FixedSizeBinaryType:
        """Parse from a string type description

        Examples:
            ```py
            s = "FixedSizeBinary[4]"
            FixedSizeBinaryType.from_string(s)
            ```
        """
        start = "FixedSizeBinary["
        end = "]"
        assert s.startswith(start)
        assert s.endswith(end)

        regex = r"^FixedSizeBinary\[(\d+)\]$"
        match = re.search(regex, s)
        assert match
        byte_width = int(match.group(1))
        return FixedSizeBinaryType(byte_width=byte_width)


class StructType(BaseDataType, BaseModel):
    type: Literal["Struct"] = "Struct"
    fields: list[Field]

    def to_arrow(self) -> pa.DataType:
        return pa.struct([field.to_arrow() for field in self.fields])

    @classmethod
    def from_arrow(cls, dtype: pa.DataType) -> StructType:
        parsed_fields: list[Field] = []
        for i in range(dtype.num_fields):
            parsed_fields.append(Field.from_arrow(dtype.field(i)))

        return StructType(fields=parsed_fields)

    @classmethod
    def from_string(cls, s: str) -> StructType:
        raise NotImplementedError

    def __str__(self) -> str:
        field_strs = [f"{field.name}:{field.type}" for field in self.fields]
        # We use a semicolon for structs to make parsing easier, since each internal
        # data type can have their own comma characters, and we don't know how many
        # struct fields will exist.
        inner_str = "; ".join(field_strs)
        return f"{self.type}[{inner_str}]"


DataType = Union[
    PrimitiveDataType,
    TimestampType,
    ListType,
    LargeListType,
    FixedSizeListType,
    FixedSizeBinaryType,
    StructType,
]


def data_type_from_arrow(dtype: pa.DataType) -> DataType:
    """Construct this from a [`pyarrow.DataType`][pyarrow.DataType]."""
    if isinstance(dtype, pa.ExtensionType):
        if dtype.extension_name == GEOARROW_NAME:
            return PrimitiveDataType.Geometry

        raise ValueError(f"Unknown extension type {dtype.extension_name}")

    if pa.types.is_null(dtype):
        return PrimitiveDataType.Null
    if pa.types.is_boolean(dtype):
        return PrimitiveDataType.Bool
    if pa.types.is_int8(dtype):
        return PrimitiveDataType.Int8
    if pa.types.is_int16(dtype):
        return PrimitiveDataType.Int16
    if pa.types.is_int32(dtype):
        return PrimitiveDataType.Int32
    if pa.types.is_int64(dtype):
        return PrimitiveDataType.Int64
    if pa.types.is_uint8(dtype):
        return PrimitiveDataType.Uint8
    if pa.types.is_uint16(dtype):
        return PrimitiveDataType.Uint16
    if pa.types.is_uint32(dtype):
        return PrimitiveDataType.Uint32
    if pa.types.is_uint64(dtype):
        return PrimitiveDataType.Uint64
    if pa.types.is_float16(dtype):
        return PrimitiveDataType.Float16
    if pa.types.is_float32(dtype):
        return PrimitiveDataType.Float32
    if pa.types.is_float64(dtype):
        return PrimitiveDataType.Float64
    if pa.types.is_date32(dtype):
        return PrimitiveDataType.Date32
    if pa.types.is_date64(dtype):
        return PrimitiveDataType.Date64
    if pa.types.is_string(dtype):
        return PrimitiveDataType.String
    if pa.types.is_large_string(dtype):
        return PrimitiveDataType.LargeString
    if pa.types.is_binary(dtype):
        return PrimitiveDataType.Binary
    if pa.types.is_large_binary(dtype):
        return PrimitiveDataType.LargeBinary
    if pa.types.is_timestamp(dtype):
        return TimestampType.from_arrow(dtype)
    if pa.types.is_list(dtype):
        return ListType.from_arrow(dtype)
    if pa.types.is_large_list(dtype):
        return LargeListType.from_arrow(dtype)
    if pa.types.is_fixed_size_list(dtype):
        return FixedSizeListType.from_arrow(dtype)
    if pa.types.is_fixed_size_binary(dtype):
        return FixedSizeBinaryType.from_arrow(dtype)
    if pa.types.is_struct(dtype):
        return StructType.from_arrow(dtype)

    # Hack: for now, cast decimal data types to float64
    if pa.types.is_decimal(dtype):
        return PrimitiveDataType.Float64

    warnings.warn(
        FusedDefaultWarning(f"unimplemented type {dtype}; defaulting to String")
    )
    # Cast default type to string??
    return PrimitiveDataType.String


def data_type_from_string(s: str) -> DataType:
    """Construct DataType from a string representation"""
    if s in ("str", "string"):
        return PrimitiveDataType.String

    try:
        return PrimitiveDataType(s)

    except ValueError:
        pass

    # If the string is not found in the PrimitiveDataType enum, first try to parse it
    # using numpy
    try:
        numpy_dtype = np.dtype(s)
        return PrimitiveDataType.from_numpy_dtype(np.dtype(numpy_dtype))
    except TypeError:
        pass

    if s.startswith("Timestamp"):
        return TimestampType.from_string(s)

    if s.startswith("List"):
        return ListType.from_string(s)

    if s.startswith("LargeList"):
        return LargeListType.from_string(s)

    if s.startswith("FixedSizeList"):
        return FixedSizeListType.from_string(s)

    if s.startswith("FixedSizeBinary"):
        return FixedSizeBinaryType.from_string(s)

    if s.startswith("Struct"):
        return StructType.from_string(s)

    raise ValueError(f"unimplemented type: {s}")


class Field(BaseModel):
    """A description of a single column in a table."""

    name: str
    """The name of the column."""

    type: DataType
    """The [data type][fused.models.DataType] of this column."""

    nullable: bool = True
    """Whether the column can have null values."""

    metadata: dict[str, str] | None = None
    """Optional metadata describing the column."""

    @field_validator("type")
    @classmethod
    def coerce_null_type_to_string(cls, v):
        if v == PrimitiveDataType.Null:
            return PrimitiveDataType.String

        return v

    def to_arrow(self) -> pa.Field:
        """Convert this to a [`pyarrow.Field`][pyarrow.Field]."""

        return pa.field(
            name=self.name,
            type=self.type.to_arrow(),
            nullable=self.nullable,
            metadata=self.metadata,
        )

    @classmethod
    def from_arrow(cls, field: pa.Field) -> Field:
        """Construct this from a [`pyarrow.Field`][pyarrow.Field]."""

        parsed_metadata: dict[str, str] = {}
        if field.metadata:
            for key, value in field.metadata.items():
                parsed_metadata[key.decode()] = value.decode()

        return cls(
            name=field.name,
            type=data_type_from_arrow(field.type),
            nullable=field.nullable,
            metadata=parsed_metadata or None,
        )

    @field_validator("type", mode="before")
    @classmethod
    def coerce_type(cls, v):
        import pandas as pd

        # For numpy, most people pass around numpy _classes_ as the type
        # Note: passing a class instance to issubclass will error
        if inspect.isclass(v) and issubclass(v, np.generic):
            return PrimitiveDataType.from_numpy_type_class(v)

        # This is rare, it would be a scalar instance of a numpy type
        if isinstance(v, np.generic):
            return PrimitiveDataType.from_numpy_type(v)

        if isinstance(v, np.dtype):
            return PrimitiveDataType.from_numpy_dtype(v)

        if isinstance(v, pd.api.extensions.ExtensionDtype):
            return PrimitiveDataType.from_pandas_type(v)

        if any(v is python_type for python_type in [float, int, str, bool]):
            return PrimitiveDataType.from_python_type(v)

        if isinstance(v, str):
            return data_type_from_string(v)

        return v


# Note: this section needs to happen _after_ all recursive references have been defined
Field.model_rebuild()
TimestampType.model_rebuild()
ListType.model_rebuild()
LargeListType.model_rebuild()
FixedSizeListType.model_rebuild()
FixedSizeBinaryType.model_rebuild()
StructType.model_rebuild()


class Schema(BaseModel):
    """A description of a table schema.

    !!! info

        This schema mirrors the definition of a [`pyarrow.Schema`][pyarrow.Schema]
        object.
    """

    fields: list[Field] = PydanticField(default_factory=list)
    """A list of [`fields`][fused.models.Field] describing the columns in this table.

    The ordering of the field descriptors must match the actual ordering of the columns
    in the table.
    """

    metadata: dict[str, str] | None = None
    """Optional metadata describing the table."""

    def to_arrow(self) -> pa.Schema:
        """Convert this to a [`pyarrow.Schema`][pyarrow.Schema]."""

        return pa.schema([field.to_arrow() for field in self.fields], self.metadata)

    @classmethod
    def empty(cls) -> Schema:
        """Returns a Schema with no fields."""
        return cls(fields=[])

    @property
    def is_empty(self) -> bool:
        """True if this schema has no fields and no metadata."""
        return not len(self.fields) and not self.metadata

    @classmethod
    def from_arrow(cls, schema: pa.Schema) -> Schema:
        """Construct this from a [`pyarrow.Schema`][pyarrow.Schema]."""

        parsed_fields: list[Field] = []
        for field in schema:
            parsed_fields.append(Field.from_arrow(field))

        parsed_metadata: dict[str, str] = {}
        if schema.metadata:
            for key, value in schema.metadata.items():
                parsed_metadata[key.decode()] = value.decode()

        return cls(fields=parsed_fields, metadata=parsed_metadata)

    @classmethod
    def from_dataframe(cls, df: pd.DataFrame | gpd.GeoDataFrame) -> Schema:
        """Infer a schema from a [`pandas.DataFrame`][pandas.DataFrame]."""
        if HAS_GEOPANDAS and isinstance(df, GPD_GEODATAFRAME):
            import geopandas as gpd

            # Remove the geometry column so that pa.Schema.from_pandas succeeds
            geometry_column_index = list(df.columns).index(df._geometry_column_name)
            non_geo_columns = [
                name for name in df.columns if name != df._geometry_column_name
            ]
            non_geospatial_schema = cls.from_dataframe(df[non_geo_columns])
            non_geospatial_schema.fields.insert(
                geometry_column_index,
                Field(name=df._geometry_column_name, type=PrimitiveDataType.Geometry),
            )

            from geopandas.io.arrow import _create_metadata

            # We must create the standard geo metadata, because it should be written even
            # if some chunk is not processed (e.g. in inner join.) We remove the bbox in that
            # case because it will be wrong if copied to some other chunk.
            try:
                if gpd.__version__ < "1":
                    geo_metadata = _create_metadata(df, schema_version=None)
                else:
                    # Internal API changed around GeoPandas 1.0
                    geo_metadata = _create_metadata(
                        df,
                        geometry_encoding={df._geometry_column_name: "WKB"},
                        schema_version=None,
                    )

                if "bbox" in geo_metadata["columns"][df._geometry_column_name]:
                    # May not be present if e.g. the column is empty of all values
                    del geo_metadata["columns"][df._geometry_column_name]["bbox"]

                non_geospatial_schema.metadata["geo"] = json.dumps(geo_metadata)
            except Exception as e:
                warnings.warn(FusedWarning(f"Failed to reparse geo metadata: {e}"))

            return non_geospatial_schema

        pyarrow_schema = pa.Schema.from_pandas(df, preserve_index=False)
        pyarrow_schema = pyarrow_schema.remove_metadata()

        # The fused index column should not be specified on the schema because it is an
        # index, not a column to be saved.
        fused_index_idx = pyarrow_schema.get_field_index(INDEX_COLUMN_NAME)
        if fused_index_idx >= 0:
            pyarrow_schema = pyarrow_schema.remove(fused_index_idx)

        return cls.from_arrow(pyarrow_schema)

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> Schema:
        """Create a schema from a key-value dictionary.

        Returns:
            A Schema object

        Examples:
            ```py
            columns = {
                'column1': np.uint8,
                'column2': str,
            }
            schema = Schema.from_dict(columns)
            ```

        """
        fields: list[Field] = []
        for name, dtype in d.items():
            field = Field(name=name, type=dtype)
            fields.append(field)

        return cls(fields=fields)

    @classmethod
    def from_parquet(cls, file: str | Path | IO[bytes]) -> Schema:
        """Construct a Schema from a Parquet file

        Args:
            file: Parquet file to parse as schema.

        Returns:
            Schema inferred from Parquet file
        """
        schema = pq.read_schema(file)
        return cls.from_arrow(schema)

    @classmethod
    def from_string(cls, str: str) -> Schema:
        """Parse a schema from a string, which may be in several forms.

        The string may either contain JSON, or a "command-line" form of specifying a schema
        where fields are specified as `name:type`, like so: `field1:Int32 field2:List[String]`

        ```
        column1:String column2:Float64
        ```

        Nested types can be defined using square bracket syntax, where the nested type is inside the outer type.
        For example, a list of strings:

        ```
        column1:List[String]
        ```

        A list of uint8s where each list has exactly two elements:

        ```
        "column2:FixedSizeList[Uint8, 2]"
        ```

        A list of floats with capacity of > 2^32 elements:

        ```
        column3:LargeList[Float64]
        ```

        4 bytes per element:

        ```
        column4:FixedSizeBinary[4]
        ```

        A timestamp stored as seconds:

        ```
        column5:Timestamp[s]
        ```

        A timestamp stored as seconds, with timezone:

        ```
        "column6:Timestamp[s, America/New_York]"
        ```
        """
        if str.lstrip().startswith("{"):
            return cls.model_validate_json(str)

        split = shlex.split(str, comments=True)

        if len(split) == 0:
            return Schema()

        parsed_fields: list[Field] = []
        for field in split:
            split = field.split(":", maxsplit=1)
            if len(split) == 1:
                raise NotImplementedError()

            name = split[0]
            if name.startswith("-"):
                warnings.warn(
                    FusedReservedWarning(
                        f"Field names should not begin with '-': {name}"
                    )
                )
            dtype = split[1]
            parsed_field = Field(name=name, type=dtype)  # type: ignore
            parsed_fields.append(parsed_field)

        return Schema(fields=parsed_fields)

    def to_string(self) -> str:
        """Returns a string form of this schema, which will be parseable by `from_string`."""
        # Check if any field needs to be customized beyond what the CLI form would have.
        any_field_has_metadata = any(
            [(f.metadata or not f.nullable) for f in self.fields]
        )
        if self.metadata or any_field_has_metadata:
            # The command-line form below does not support encoding metadata
            return self.model_dump_json()

        field_strs: list[str] = []
        for field in self.fields:
            field_descriptor = f"{field.name}:{field.type}"
            field_strs.append(f"{shlex.quote(field_descriptor)}")

        return " ".join(field_strs)

    def update(self, other: Schema) -> Schema:
        """Update this schema using fields from another schema.

        Fields are updated by name from the other schema. The fields in this schema will
        stay in the same order. All field names in the other schema must exist in this
        schema.

        Args:
            other: the schema with which to update this schema

        Returns:
            A new Schema object
        """
        new_metadata = {}
        left_metadata = self.metadata or {}
        right_metadata = other.metadata or {}
        new_metadata.update(left_metadata)
        left_metadata.update(right_metadata)

        left_fields = self.fields
        right_fields = other.fields

        left_field_names = {field.name for field in left_fields}
        right_field_names = {field.name for field in right_fields}
        extra_right_field_names = right_field_names.difference(left_field_names)
        if extra_right_field_names:
            error_msg = (
                f"{', '.join(extra_right_field_names)}"
                " found in right schema but not in left schema"
            )
            raise ValueError(error_msg)

        right_field_mapping = {field.name: field for field in right_fields}

        new_fields = []
        for left_field in left_fields:
            right_field = right_field_mapping.get(left_field.name)
            if right_field is not None:
                new_fields.append(right_field)
            else:
                new_fields.append(left_field)

        return Schema(fields=new_fields, metadata=new_metadata)

    def get_field(self, name: str) -> Field | None:
        """Returns the field in this schema with the specified name, or None.

        Args:
            name: The name of the field.

        Returns:
            The Field object, or None if no field with that name was found."""
        for field in self.fields:
            if field.name == name:
                return field

        return None

    @field_validator("fields")
    @classmethod
    def validate_fields(cls, v: list[Field]):
        assert len(v) == len({field.name for field in v}), "Field names must be unique"
        return v



from __future__ import annotations

import ast
import time
import warnings
from contextlib import ExitStack
from typing import Any

import fused
from fused._optional_deps import HAS_PANDAS, PD_DATAFRAME
from fused._udf.compile_v2 import compile_udf_and_run_with_kwargs
from fused._udf.state import (
    _isolate_streams,
    decorator_src_override_context,
)
from fused.models.schema import Schema
from fused.models.udf import EMPTY_UDF, AnyBaseUdf, Output, PandasOutput
from fused.models.udf._eval_result import UdfEvaluationResult
from fused.warnings import FusedWarning


def execute_against_sample(
    udf: AnyBaseUdf,
    input: list[Any],
    update_schema: bool = True,
    validate_output: bool = True,
    validate_imports: bool | None = None,
    _return_response: bool = False,
    **kwargs,
) -> UdfEvaluationResult:
    if udf is EMPTY_UDF:
        raise ValueError("Empty UDF cannot be evaluated. Use `set_udf` to set a UDF.")

    output = Output()
    try:
        # Validate import stamements correspond to valid modules
        validate_imports_whitelist(udf, validate_imports=validate_imports)

        # Run UDF
        if _return_response:
            # TODO capture output
            _output = None
            has_exception = False
            errormsg = None
            exception_class = None
            time_start = time.perf_counter()
            try:
                _output = compile_udf_and_run_with_kwargs(udf, *input, **kwargs)
            except Exception as exc:
                has_exception = True
                exception_class = exc.__class__.__name__
                errormsg = f"{exception_class}: {str(exc)}"
                # TODO proper error traceback
                # traceback.print_tb(exc.__traceback__, file=err_buf)
            time_end = time.perf_counter()
            time_taken_seconds = time_end - time_start

            return UdfEvaluationResult(
                data=_output,
                udf=udf,
                time_taken_seconds=time_taken_seconds,
                stdout=None,
                stderr=None,
                error_message=errormsg,
                has_exception=has_exception,
                exception_class=exception_class,
            )

        time_start = time.perf_counter()
        with _isolate_streams():
            _output = compile_udf_and_run_with_kwargs(udf, *input, **kwargs)
        time_end = time.perf_counter()
        time_taken_seconds = time_end - time_start

        if not validate_output:
            return _output

        if _output is not None:
            output.data = _output

    except Exception:
        raise

    new_output = _transform_output(output=output)

    if new_output is None:
        # TODO: Assumes table_schema is present, which doesn't match type
        if udf.table_schema not in [None, {}, Schema()]:
            warnings.warn(
                FusedWarning(
                    "UDF is configured with a schema but returns `None`. An empty schema was set for this execution."
                )
            )

        return UdfEvaluationResult(
            data=None,
            sidecar=None,
            udf=udf,
            table_schema={},
            time_taken_seconds=time_taken_seconds,
        )

    # Validate the dataframe after assigning it to the output variable in user_ns
    # so the user can inspect the output if anything is wrong.
    new_output.validate_data_with_schema()

    if update_schema and udf.table_schema is None:
        udf.table_schema = new_output.table_schema

    return UdfEvaluationResult(
        data=new_output.data,
        sidecar=new_output.sidecar_output,
        udf=udf,
        table_schema=new_output.table_schema,
        time_taken_seconds=time_taken_seconds,
    )


def execute_for_decorator(udf: AnyBaseUdf) -> AnyBaseUdf:
    """Evaluate a UDF for the purpose of getting the UDF object out of it."""
    # Define custom function in environment

    # This is a stripped-down version of execute_against_sample, above.

    src = udf.code

    exec_globals_locals = {"fused": fused}

    with ExitStack() as stack:
        stack.enter_context(decorator_src_override_context(src))

        # Add headers to sys.meta_path
        if udf.headers is not None:
            for header in udf.headers:
                stack.enter_context(header._register_custom_finder())

        exec(src, exec_globals_locals, exec_globals_locals)

        if udf.entrypoint not in exec_globals_locals:
            raise NameError(
                f"Could not find {udf.entrypoint}. You need to define a UDF with `def {udf.entrypoint}()`."
            )

        _fn = exec_globals_locals[udf.entrypoint]

        return _fn


def _transform_output(output: Output) -> PandasOutput:
    # TODO: Support PyArrow
    if isinstance(output, PandasOutput):
        return output

    if HAS_PANDAS and isinstance(output.data, PD_DATAFRAME):
        # Force all column names to be strings
        output.data.columns = [str(x) for x in output.data.columns]

        # Note that we don't pass a _new_ schema argument here
        # - If the user defined a schema onto output.table_schema, that will be
        #   preserved.
        # - If the user did not define output.table_schema, then output.table_schema
        #   will be None, and PandasOutput will default to a schema inferred from
        #   the DataFrame.
        return PandasOutput(
            data=output.data,
            table_schema=output.table_schema,
            sidecar_output=output.sidecar_output,
            skip_fused_index_validation=output.skip_fused_index_validation,
        )
    elif output.data is None:
        return None

    raise TypeError(f"Unexpected result type {type(output.data)}")


def validate_imports_whitelist(udf: AnyBaseUdf, validate_imports: bool | None = None):
    # Skip import validation if the option is set
    if not fused.options.default_validate_imports and validate_imports is not True:
        return

    # Skip import validation if not logged in
    if not fused.api.AUTHORIZATION.is_configured():
        return

    from fused._global_api import get_api

    # Get the dependency whitelist from the cached API endpoint
    api = get_api()
    package_dependencies = api.dependency_whitelist()

    # Initialize a list to store the import statements
    import_statements = []

    # Parse the source code into an AST
    tree = ast.parse(udf.code)

    # Traverse the AST to find import statements
    for node in ast.walk(tree):
        if isinstance(node, ast.Import):
            for alias in node.names:
                import_statements.append(alias.name)
        elif isinstance(node, ast.ImportFrom):
            module_name = node.module
            import_statements.append(module_name)

    # Check for unavailable modules
    header_modules = [header.module_name for header in udf.headers]
    fused_modules = ["fused"]  # assume fused is always available
    available_modules = (
        list(package_dependencies["dependency_whitelist"].keys())
        + header_modules
        + fused_modules
    )
    unavailable_modules = []
    for import_statement in import_statements:
        if import_statement.split(".", 1)[0] not in available_modules:
            unavailable_modules.append(import_statement)

    if unavailable_modules:
        raise ValueError(
            f"The following imports in the UDF might not be available: {repr(unavailable_modules)}. Please check the UDF headers and imports and try again."
        )

    # TODO: check major versions for some packages



import warnings
from io import BytesIO
from typing import TYPE_CHECKING, Union

from fused.warnings import FusedImportWarning

if TYPE_CHECKING:
    import geopandas as gpd
    import pandas as pd


def parquet_to_df(b: bytes) -> Union["pd.DataFrame", "gpd.GeoDataFrame"]:
    try:
        import pandas as pd

        try:
            import geopandas as gpd

            return gpd.read_parquet(BytesIO(b))

        except:  # noqa: E722
            df = pd.read_parquet(BytesIO(b))

            if (
                "geometry" in df.columns
                and bytes in df["geometry"].apply(type).unique()
            ):
                warnings.warn(
                    FusedImportWarning(
                        "`geopandas` package is not installed so geometries are displayed as bytes instead of parsed shapes"
                    )
                )

            return df

    except ImportError:
        raise ModuleNotFoundError(
            "`pandas` package is not installed. Please install `pandas` to run this UDF."
        )



try:
    import pandas as pd

    HAS_PANDAS = True
    PD_DATAFRAME = pd.DataFrame
    PD_SERIES = pd.Series
    PD_INDEX = pd.Index
    PD_HASH_OBJECT_FN = pd.util.hash_pandas_object
except ImportError:
    HAS_PANDAS = False
    PD_DATAFRAME = None
    PD_SERIES = None
    PD_INDEX = None
    PD_HASH_OBJECT_FN = None

try:
    import geopandas as gpd

    HAS_GEOPANDAS = True
    GPD_GEODATAFRAME = gpd.GeoDataFrame
except ImportError:
    HAS_GEOPANDAS = False
    GPD_GEODATAFRAME = None

try:
    import mercantile

    HAS_MERCANTILE = True
    MERCANTILE_TILE = mercantile.Tile
except ImportError:
    HAS_MERCANTILE = False
    MERCANTILE_TILE = None

try:
    import shapely

    HAS_SHAPELY = True
    SHAPELY_GEOMETRY = shapely.Geometry
    SHAPELY_POLYGON = shapely.Polygon
except ImportError:
    HAS_SHAPELY = False
    SHAPELY_GEOMETRY = None
    SHAPELY_POLYGON = None

try:
    import fsspec  # noqa: F401

    HAS_FSSPEC = True
except ImportError:
    HAS_FSSPEC = False



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from fused.api import FusedAPI, FusedDockerAPI


API: FusedAPI | FusedDockerAPI | None = None

# We expect this to be set after import
API_CLASS: type[FusedAPI] | None = None


def get_api(**api_kwargs) -> FusedAPI | FusedDockerAPI:
    global API
    if API is not None:
        return API

    if API_CLASS is not None:
        API = API_CLASS(**api_kwargs)
        return API

    raise ValueError("Internal error: Global API_CLASS was not set.")


def set_api(api: FusedAPI | FusedDockerAPI) -> None:
    global API
    API = api


def set_api_class(api_class: type[FusedAPI]) -> None:
    global API_CLASS
    API_CLASS = api_class


def reset_api():
    global API
    API = None



from __future__ import annotations

import ast
import difflib
import json
import re
import warnings
from contextlib import ExitStack
from copy import deepcopy
from enum import Enum
from functools import cached_property
from io import IOBase
from pathlib import PurePath
from textwrap import dedent, indent
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    Literal,
    Sequence,
)
from uuid import UUID

import requests
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    RootModel,
    field_validator,
    model_validator,
)
from pydantic_core.core_schema import ValidationInfo
from requests import HTTPError
from typing_extensions import Self, override

from fused._formatter.udf import fused_header_repr, fused_udf_repr
from fused._options import options as OPTIONS
from fused.models.base import UserMetadataType
from fused.models.udf.header import Header
from fused.warnings import FusedDeprecationWarning, FusedWarning

from .._codegen import (
    extract_parameters,
    stringify_headers,
    stringify_named_params,
    structure_params,
)
from .._inplace import _maybe_inplace

if TYPE_CHECKING:
    from fused.models.api import UdfAccessToken, UdfAccessTokenList

from fused._global_api import get_api

METADATA_FUSED_ID = "fused:id"
METADATA_FUSED_SLUG = "fused:slug"
METADATA_FUSED_EXPLORER_TAB = "fused:explorerTab"


class HeaderSequence(RootModel[Sequence[Header]]):
    def _repr_html_(self) -> str:
        return fused_header_repr(self)

    def __iter__(self):
        return iter(self.root)

    def __getitem__(self, item):
        return self.root[item]

    def __setitem__(self, item, value):
        self.root[item] = value

    def __len__(self):
        return len(self.root)


class AttrDict(dict):
    """Dictionary where keys can also be accessed as attributes"""

    def __getattribute__(self, __name: str) -> Any:
        try:
            return super().__getattribute__(__name)
        except AttributeError:
            if __name in self:
                return self[__name]
            else:
                raise

    def __dir__(self) -> Iterable[str]:
        return self.keys()


class CompiledAttrs(AttrDict):
    """Dictionary that stores compiled values where keys can also be accessed as attributes"""

    def __init__(self, header: Header = None, do_import: bool = False):
        self.do_import = do_import
        if header:
            self.header = header
            if do_import:
                from fused._udf.state import noop_decorators_context

                with noop_decorators_context(True):
                    super().__init__(header._exec())

    def _do_import(self):
        from fused._udf.state import noop_decorators_context

        with noop_decorators_context(True):
            self.update(self.header._exec())

    def __deepcopy__(self, memo=None):
        # Deepcopy can happen when pydantic model_copy(deep=True) is called on UDF through _maybe_inplace
        # Note: IPython added additional values to __builtins__ globals which caused further issues during this process.
        cls = type(self)
        return cls(deepcopy(self.header, memo), self.do_import)


class UdfType(str, Enum):
    GEOPANDAS_V2 = "geopandas_v2"


class BaseUdf(BaseModel):
    name: str | None = None
    type: Literal[UdfType.GEOPANDAS_V2]
    code: str
    headers: HeaderSequence | Sequence[Header] = Field(default_factory=list)
    metadata: UserMetadataType = None
    _globals: CompiledAttrs | None = None
    _import_globals: bool = True

    @model_validator(mode="after")
    def _process_import(self, info: ValidationInfo) -> Self:
        self._import_globals = (
            info.context.get("import_globals") if info.context else self._import_globals
        )
        if not OPTIONS.never_import and self._import_globals:
            if not self.code:
                # This is to work around initializing EMPTY_UDF
                self._globals = CompiledAttrs()
            else:
                with ExitStack() as stack:
                    if self.headers is not None:
                        for header in self.headers:
                            stack.enter_context(header._register_custom_finder())

                    module_name = f"{self.name or 'udf'}"
                    self._globals = CompiledAttrs(
                        Header(
                            module_name=module_name,
                            source_code=self.code,
                            source_file=f"{module_name}.py",
                        ),
                        not OPTIONS.never_import,
                    )

        return self

    def import_globals(self):
        self._globals._do_import()

    def __deepcopy__(self, memo=None):
        # deep copying the utils module might fail (in pickling)
        try:
            del self._cached_utils
        except AttributeError:
            pass
        return super().__deepcopy__(memo)

    def __eq__(self, other):
        # Quick check of equality by only checking types and public variables.
        # Pydantic __eq__ compare private attributes which fails existing equality checks.

        if not isinstance(other, self.__class__):
            return NotImplemented

        # Only compare public attributes as _globals contains un-comparable values.
        return self.__dict__ == other.__dict__

    def __getattr__(self, item: str):
        # Any private attribute access gets passed here (pydantic behavior)
        # Attribute access for those not found on the model also gets passed here (python behavior)

        method_attrs = {"utils", "_cached_utils"}
        # Ensure we prioritize access to the model's private attributes and method attributes
        if item in (self.__private_attributes__.keys() | method_attrs):
            return super().__getattr__(item)

        # This access takes advantage of the above private attributes check or else we'd be in an infinite loop.
        _g = self._globals
        if _g is None:
            raise AttributeError(
                "Accessing attributes on this UDF was disabled when it was loaded."
            )
        try:
            return getattr(_g, item)
        except AttributeError:
            raise AttributeError(
                f"{self.__class__.__name__!r} object has no attribute {item!r}"
            )

    @override
    def __repr_args__(self):
        metadata = self.metadata or {}
        yield "id", metadata.get("fused:id")
        yield "name", self.name
        yield "description", metadata.get("fused:description")

        # Define the metadata keys to potentially include
        optional_metadata_keys = {
            "fused:udfType": "udf_type",
            "fused:slug": "slug",
            "fused:gitRepo": "git_repo",
            "fused:gitLastModified": "last_modified",
        }

        for meta_key, repr_key in optional_metadata_keys.items():
            value = metadata.get(meta_key)
            if value:
                yield repr_key, value

        yield "code", self.code

    @override
    def __repr__(self) -> str:
        args_str = self.__repr_str__(",\n  ")
        return f"{self.__repr_name__()}(\n  {args_str}\n)"

    @override
    def __str__(self) -> str:
        return self.__repr__()

    @field_validator("headers", mode="before")
    @classmethod
    def _process_headers(cls, headers):
        processed_headers = []
        if headers is not None:
            for header in headers:
                if isinstance(header, str):
                    module_name = PurePath(header).name.split(".", maxsplit=1)[0]
                    processed_headers.append(
                        Header.from_code(module_name=module_name, source_file=header)
                    )
                elif (
                    isinstance(header, dict)
                    and header.get("source_file")
                    and not header.get("source_code")
                ):  # headers = {'source_file': 'header.py', 'module_name': 'header'}
                    source_file = header["source_file"]
                    module_name = header.get(
                        "module_name",
                        PurePath(source_file).name.split(".", maxsplit=1)[0],
                    )
                    processed_headers.append(
                        Header.from_code(
                            module_name=module_name,
                            source_file=source_file,
                        )
                    )
                else:
                    processed_headers.append(header)

        return processed_headers

    @field_validator("code", mode="before")
    @classmethod
    def udf_is_str(cls, v):
        if isinstance(v, IOBase):
            # Handle passing in a file as the code
            v.seek(0)
            data = v.read()
            if isinstance(data, str):
                return data
            elif isinstance(data, bytes):
                return data.decode("utf-8")
            else:
                raise ValueError("Expected string or bytes from the file")

        return v

    @field_validator("code", mode="after")
    @classmethod
    def code_valid_max_cache_age(cls, code: str):
        # Regex to find @fused.udf decorator and capture the cache_max_age value
        # It handles single or double quotes and optional spacing.
        # re.DOTALL allows '.' to match newline characters if the decorator spans multiple lines.
        match: re.Match[str] | None = re.search(
            r"@fused\.udf\(.*?cache_max_age\s*=\s*[\"']([^\"']+)['\"].*?\)",
            code,
            re.DOTALL,
        )

        if match:
            cache_age_str = match.group(1)
            # validate the cache_age_str to check it follows the standard format
            valid_value = re.match(r"(\d+)([smhd])$", cache_age_str.strip().lower())
            if not valid_value:
                raise ValueError(
                    f"Invalid `cache_max_age` parameter '{cache_age_str}': Use a number followed by one of 's' (seconds), 'm' (minutes), 'h' (hours), or 'd' (days)."
                )

        return code

    @classmethod
    def from_gist(cls, gist_id: str):
        """Create a Udf from a GitHub gist."""
        # TODO: if a versioned gist, taking the last / won't work as that will be the
        # version, not the gist id.
        if "/" in gist_id:
            gist_id = gist_id.split("/")[-1]

        url = f"https://api.github.com/gists/{gist_id}"
        r = requests.get(url, timeout=OPTIONS.request_timeout)
        r.raise_for_status()

        files: dict = r.json()["files"]

        obj = {}
        # TODO: Update expected files schema
        assert "code.py" in files.keys()
        assert "parameters.json" in files.keys()

        obj["code"] = _fetch_gist_content(files, "code.py")

        if "table_schema.json" in files.keys():
            table_schema_text = _fetch_gist_content(files, "parameters.json")
            obj["table_schema"] = json.loads(table_schema_text)

        parameters_text = _fetch_gist_content(files, "parameters.json")
        parameters = json.loads(parameters_text)
        obj.update(parameters)

        return cls.model_validate(obj)

    # TODO: Rename to to_fused when deprecated features in existing to_fused are removed
    def _to_fused_v2(
        self,
        overwrite: bool | None = None,
    ):
        """
        Save this UDF on the Fused service.

        Args:
            overwrite: If True, overwrite existing remote UDF with the UDF object.
        """
        api = get_api()
        self_id = self._get_metadata_safe(METADATA_FUSED_ID)
        remote_udf = self._maybe_udf_by_name(self.name)
        backend_id = self_id
        old_metadata = None
        if remote_udf:
            remote_id = remote_udf._get_metadata_safe(METADATA_FUSED_ID)
            # Handle legacy use case of default overwriting if same udf
            if self_id == remote_id and overwrite is None:
                overwrite = True
            if overwrite:
                backend_id = remote_id

        # Ensures some metadata values are not serialized as it can lead to stale data upon loading.
        if self.metadata:
            old_metadata = self.metadata.copy() if self.metadata else {}
            self._delete_metadata_safe(METADATA_FUSED_ID)
            self._delete_metadata_safe(METADATA_FUSED_SLUG)

        try:
            result = api.save_udf(udf=self, slug=self.name, id=backend_id)
        except HTTPError as e:
            # If already exists:
            if e.response.status_code == 409:
                diff_lines = difflib.unified_diff(
                    (remote_udf.code + "\n").splitlines(keepends=True),
                    (self.code + "\n").splitlines(keepends=True),
                    fromfile="remote",
                    tofile="local",
                )
                raise ValueError(
                    f"{e.args[0]['detail']}. A diff of the code has been provided. Use overwrite=True if you want to replace the remote UDF with the local one.\n\n"
                    + "".join(diff_lines)
                )
            if old_metadata:
                self.metadata = old_metadata
            raise

        new_id = result["id"]
        self._set_metadata_safe(METADATA_FUSED_ID, new_id)
        self._set_metadata_safe(METADATA_FUSED_SLUG, self.name)
        return self

    def to_fused(
        self,
        slug: str | None = ...,
        over_id: str | UUID | None = None,
        as_new: bool | None = None,
        inplace: bool = True,
        overwrite: bool | None = None,
    ):
        """
        Save this UDF on the Fused service.

        Args:
            slug: ID to refer to this UDF as in URLs.
            over_id: ID to save the UDF over.
            as_new: If True, force saving this UDF as new.
            inplace: If True (default), update the UDF object with the new saved ID.
            overwrite: If True, overwrite existing remote UDF with the UDF object.
        """

        if slug is not Ellipsis:
            warnings.warn(
                "'slug' parameter is deprecated. The name of the UDF will be used to generate the slug.",
                FusedDeprecationWarning,
                stacklevel=2,
            )

        if over_id:
            warnings.warn(
                "'over_id' parameter is deprecated.",
                FusedDeprecationWarning,
                stacklevel=2,
            )
        if as_new:
            warnings.warn(
                "'as_new' parameter is deprecated.",
                FusedDeprecationWarning,
                stacklevel=2,
            )

        if inplace is False:
            warnings.warn(
                "'inplace' parameter is deprecated.",
                FusedDeprecationWarning,
                stacklevel=2,
            )

        # New Flow
        if not any((slug is not Ellipsis, over_id, as_new, inplace is False)):
            return self._to_fused_v2(overwrite)

        backend_id = (
            str(over_id) if over_id else self._get_metadata_safe(METADATA_FUSED_ID)
        )
        if backend_id is None or as_new:
            assert as_new is not False, (
                "Cannot detect ID to save over, so as_new cannot be False."
            )
            backend_id = None

        ret = _maybe_inplace(self, inplace)

        if slug is Ellipsis:
            # If the user didn't specify a name to save as, determine the name
            # to save as here.
            slug = ret._get_metadata_safe(METADATA_FUSED_SLUG)
            if not slug:
                # No metadata-determined name, so use the regular name of the
                # UDF.
                slug = ret.name

        if slug and slug != ret.name:
            # If we are setting the name of the UDF, it is important to set the
            # name on the UDF body itself, because that is what Workbench will read
            ret.name = slug

        api = get_api()
        result = api.save_udf(
            udf=ret,
            slug=slug,
            id=backend_id,
        )
        new_id = result["id"]

        ret._set_metadata_safe(METADATA_FUSED_ID, new_id)
        ret._set_metadata_safe(METADATA_FUSED_SLUG, slug)
        return ret

    def _maybe_udf_by_name(self, slug: str):
        from .udf import load_udf_from_response_data

        api = get_api()
        try:
            remote_udf_data = api._get_udf(slug)
        except HTTPError:
            return None

        udf = load_udf_from_response_data(remote_udf_data)
        return udf

    def delete_saved(self, inplace: bool = True):
        from fused._global_api import get_api

        backend_id = self._get_metadata_safe(METADATA_FUSED_ID)
        assert backend_id is not None, "No saved UDF ID found in metadata."

        api = get_api()
        api.delete_saved_udf(
            id=backend_id,
        )

        ret = _maybe_inplace(self, inplace)
        # ret.metadata must be non None because we read the backend ID from there
        ret.metadata.pop(METADATA_FUSED_ID)
        return ret

    def delete_cache(self):
        backend_id = self._get_metadata_safe(METADATA_FUSED_ID)
        assert backend_id is not None, "No saved UDF ID found in metadata."

        api = get_api()
        api.delete_cache(backend_id)

        return self

    def create_access_token(
        self,
        *,
        client_id: str | Ellipsis | None = ...,
        public_read: bool | None = None,
        access_scope: str | None = None,
        cache: bool = True,
        metadata_json: dict[str, Any] | None = None,
        enabled: bool = True,
    ) -> UdfAccessToken:
        from fused._global_api import get_api

        # If there is a backend ID, locate this UDF by the backend ID rather than by its name
        backend_id = self._get_metadata_safe(METADATA_FUSED_ID)
        assert backend_id is not None, (
            "No saved UDF ID found in metadata. Save the UDF to Fused first with `.to_fused` and then create the access token."
        )

        api = get_api()
        return api.create_udf_access_token(
            udf_id=backend_id,
            client_id=client_id,
            public_read=public_read,
            access_scope=access_scope,
            cache=cache,
            metadata_json=metadata_json,
            enabled=enabled,
        )

    def get_access_tokens(self) -> UdfAccessTokenList:
        from fused._global_api import get_api
        from fused.models.api import UdfAccessTokenList

        # If there is a backend ID, locate this UDF by the backend ID rather than by its name
        backend_id = self._get_metadata_safe(METADATA_FUSED_ID)
        assert backend_id is not None, (
            "No saved UDF ID found in metadata. Save the UDF to Fused first with `.to_fused` and then create the access token."
        )

        api = get_api()
        all_tokens = api.get_udf_access_tokens(max_requests=None)

        return UdfAccessTokenList(
            [token for token in all_tokens if token.udf_id == backend_id]
        )

    def _repr_html_(self) -> str:
        return fused_udf_repr(self)

    def _get_metadata_safe(self, key: str, default: Any | None = None) -> str | None:
        if self.metadata is not None:
            return self.metadata.get(key, default)
        else:
            return None

    def _set_metadata_safe(self, key: str, value: Any):
        if self.metadata is None:
            self.metadata = {}

        self.metadata[key] = value

    def _delete_metadata_safe(self, key: str):
        try:
            del self.metadata[key]
        except KeyError:
            pass

    def _generate_code(
        self, include_imports=True, headerfile=False
    ) -> tuple[str, Sequence[str]]:
        def _extract_fn(src: str) -> ast.FunctionDef | ast.AsyncFunctionDef:
            # TODO: handle Header objects in header param
            parsed_ast = ast.parse(src)

            # Find all function definitions in the AST
            function_defs = [
                node
                for node in ast.walk(parsed_ast)
                if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef))
            ]

            # first function (assume it's the target function)
            return function_defs[0]

        def _extract_fn_body(
            function_def: ast.FunctionDef | ast.AsyncFunctionDef, src: str
        ) -> str:
            target_function_body = function_def.body
            # Reconstruct the source code of the function body
            line_start = target_function_body[0].lineno - 1
            line_end = target_function_body[-1].end_lineno
            target_function_body_lines = [
                line for line in src.splitlines()[line_start:line_end]
            ]
            target_function_body_str = "\n".join(target_function_body_lines)

            return target_function_body_str

        # Derive parameters
        positional_parameters, named_parameters = extract_parameters(self.code)

        _params_fn_original = positional_parameters + stringify_named_params(
            named_parameters
        )
        # String: Imports
        str_imports = (
            "\n".join(
                [
                    "import fused",
                    "from fused.models.udf import Header",
                ]
            )
            + "\n\n"
        )
        # String: UDF header - Replace Header with file reference
        header_files = []
        processed_headers = []
        # TODO: do this conversion as part of header attribute
        for header in self.headers:
            if isinstance(header, Header):
                if headerfile:
                    filename = header.module_name + ".py"
                    processed_headers.append(filename)

                else:
                    processed_headers.append(header)
                # Create a file magic string
                header_files.append(header._generate_cell_code())

            else:
                processed_headers.append(header)

        _headers = (
            stringify_headers(processed_headers) if headerfile else processed_headers
        )
        _schema = repr(self.table_schema.to_string()) if self.table_schema else '""'
        params_decorator = [f"\n    headers={_headers}", f"\n    schema={_schema}\n"]
        str_udf_header = (
            f"@fused.udf({structure_params(params_decorator, separator=',')})"
        )
        # String: Function header
        fn = _extract_fn(self.code)
        str_async = "async " if isinstance(fn, ast.AsyncFunctionDef) else ""
        str_fn_header = f"{str_async}def {self.entrypoint}({structure_params(_params_fn_original)}):"
        # String: Function body
        str_fn_body = dedent(_extract_fn_body(fn, src=self.code))

        str_udf = f"""
{str_udf_header}
{str_fn_header}
{indent(str_fn_body, " " * 4)}
"""
        if include_imports:
            str_udf = str_imports + str_udf
        return str_udf.strip("\n"), header_files

    def render(self):
        from IPython import get_ipython
        from IPython.core.inputsplitter import IPythonInputSplitter

        # Get the current IPython instance.
        ipython = get_ipython()
        if ipython is None:
            raise RuntimeError("This function can only be used in a Jupyter Notebook.")

        # Create an instance of IPythonInputSplitter.
        splitter = IPythonInputSplitter()

        # Generate code string and split into lines.
        lines = self._generate_code()[0].strip().split("\n")

        # Set the content of the subsequent cell with.
        ipython.set_next_input(splitter.transform_cell("\n".join(lines)))

    @property
    def utils(self):
        return self._cached_utils

    @cached_property
    def _cached_utils(self):
        if len(self.headers) == 0:
            raise ValueError("UDF does not have a utils module")
        if len(self.headers) > 1:
            raise ValueError("UDF has multiple header modules")
        if self.headers[0].module_name != "utils":
            warnings.warn(
                FusedWarning(
                    f"Accessing header module {self.headers[0].module_name} under the name utils"
                )
            )
        # TODO: Even though this might have already been evaluated, we have to evaluate it again now
        # It is at least cached

        vals = self.headers[0]._exec()

        return AttrDict(vals)

    model_config = ConfigDict(exclude={"utils"})

    def __getstate__(self):
        # Pickling the utils will fail (module is not pickleable)
        try:
            del self._cached_utils
        except AttributeError:
            pass

        state = super().__getstate__()
        # Remove the _globals and _nested_callable attributes from the state
        # (modules and inline functions are not pickleable)
        state["__pydantic_private__"].pop("_globals", None)
        state["__pydantic_private__"].pop("_nested_callable", None)
        return state

    def __setstate__(self, state):
        super().__setstate__(state)
        # Recreate the _globals and _nested_callable attributes
        self._globals = None
        self._nested_callable = None
        # validate to ensure we run _process_import
        self.model_validate(self)


def _fetch_gist_content(gist_files_dict: dict, fname: str) -> str:
    gist_data = gist_files_dict[fname]

    if gist_data["truncated"]:
        full_url = gist_data["raw_url"]
        full_r = requests.get(full_url, timeout=OPTIONS.request_timeout)
        full_r.raise_for_status()
        return full_r.text
    else:
        return gist_data["content"]



import os
import tempfile
import warnings
from pathlib import Path
from typing import Dict, List, Literal, Optional, Tuple, Union
from urllib.parse import urlparse

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    StrictBool,
    StrictStr,
    field_serializer,
    field_validator,
)

from ._global_api import reset_api
from .warnings import FusedIgnoredWarning, FusedImportWarning

DEV_DEFAULT_BASE_URL = "http://localhost:8783/v1"
PROD_DEFAULT_BASE_URL = "https://www.fused.io/server/v1"
STAGING_DEFAULT_BASE_URL = "https://www-staging.fused.io/server/v1"
UNSTABLE_DEFAULT_BASE_URL = "https://unstable.fused.io/server/v1"
DEV_PR_DEFAULT_BASE_URL = "https://{}.dev.fusedlabs.io/server/v1"


OPTIONS_PATH = Path("~/.fused/settings.toml").expanduser()
"""First choice for where to find settings"""
OPTIONS_JSON_PATH = Path("~/.fused/settings.json").expanduser()
"""Second choice for where to find settings"""

StorageStr = Literal["auto", "mount", "local"]


class OptionsBaseModel(BaseModel):
    def __dir__(self) -> List[str]:
        # Provide method name lookup and completion. Only provide 'public'
        # methods.
        # This enables autocompletion
        # Pydantic methods to remove in __dir__
        PYDANTIC_METHODS = {
            "Config",
            "construct",
            "copy",
            "from_orm",
            "json",
            "parse_file",
            "parse_obj",
            "schema",
            "schema_json",
            "update_forward_refs",
            "validate",
            "model_validate",
            "model_dump_json",
        }

        normal_dir = {
            name
            for name in dir(type(self))
            if (not name.startswith("_") and name not in PYDANTIC_METHODS)
        }
        pydantic_fields = set(self.model_fields.keys())
        return sorted(normal_dir | pydantic_fields)

    def _repr_html_(self) -> str:
        # Circular import because the repr needs the options
        from fused._formatter.formatter_options import fused_options_repr

        return fused_options_repr(self)


class AuthOptions(OptionsBaseModel):
    authorize_url: str = "https://dev-tjcykxcetrz6bps6.us.auth0.com/authorize"
    """The authorize URL is used for the initial login flow. This is intended to be opened in
    the user's web browser for them to sign in."""

    oauth_token_url: str = "https://dev-tjcykxcetrz6bps6.us.auth0.com/oauth/token"
    """The token url is used for programmatic access to generate access and refresh tokens."""

    logout_url: str = "https://dev-tjcykxcetrz6bps6.us.auth0.com/oidc/logout"

    # The client id, client secret, and audience identifies a specific application and API
    client_id: str = "CXiwKZQmmyo0rqXZY7pzBgfsF7AL2A9l"
    client_secret: str = (
        "FVNz012KgNmqITYnCCOM8Q1Nt81W_DO4SeCRgVsftREKTWpzZU522nia5TdSNv8h"
    )
    audience: str = "fused-python-api"

    local_redirect_url: str = "http://localhost:3000"
    """This redirect uri is passed to the authorize URL as a url parameter. This localhost
    uri is used to intercept the "code" generated from the authorization"""

    scopes: List[str] = ["openid", "email", "name", "offline_access"]
    """The offline_access scope is necessary to be able to fetch refresh tokens
    The other scopes are useful to access identifying information in the retrieved JWT"""

    credentials_path: str = "~/.fused/credentials"
    """The path where the refresh token is saved on disk. Will be user-expanded to resolve ~."""


class ShowOptions(OptionsBaseModel):
    """Options for showing debug information"""

    open_browser: Optional[StrictBool] = None
    """Whether to open a local browser window for debug information"""
    show_widget: Optional[StrictBool] = None
    """Whether to show debug information in an IPython widget"""

    enable_tqdm: StrictBool = True
    """Whether to show tqdm-based progress information"""


def _cache_directory() -> Path:
    return get_cache_dir("auto")


def _data_directory() -> Path:
    return get_data_dir("auto")


def get_data_dir(storage: StorageStr = "auto") -> Path:
    base_path = get_writable_dir(storage)
    # Keep consistency with job2 data_path where return can be /mount/tmp or /tmp
    # TODO: Maybe unify to single common dir - /mount/data and /tmp/data
    if "mount" in str(base_path):
        return base_path / "tmp"
    else:
        return base_path


def get_cache_dir(storage: StorageStr = "auto") -> Path:
    base_path = get_writable_dir(storage)
    return base_path / "cached_data"


def get_writable_dir(storage: StorageStr = "auto") -> Path:
    mount_path = Path("/mount")
    temp_dir = Path(tempfile.gettempdir())

    if storage == "auto":
        # Cache in mounted drive if available & writable, else cache in /tmp
        if is_path_writable(mount_path):
            base_path = mount_path
        else:
            base_path = temp_dir
    elif storage == "mount":
        base_path = mount_path
    elif storage == "local":
        base_path = temp_dir
    else:
        raise ValueError(storage)

    return base_path


def is_path_writable(p: Path) -> bool:
    return p.exists() and os.access(p, os.W_OK)


class Options(OptionsBaseModel):
    base_url: str = PROD_DEFAULT_BASE_URL
    """Fused API endpoint"""

    auth: AuthOptions = Field(default_factory=AuthOptions)
    """Options for authentication."""

    show: ShowOptions = Field(default_factory=ShowOptions)
    """Options for object reprs and how data are shown for debugging."""

    max_workers: int = 16
    """Maximum number of threads, when multithreading requests"""

    request_timeout: Union[Tuple[float, float], float, None] = 120
    """Request timeout for the Fused service

    May be set to a tuple of connection timeout and read timeout"""

    request_max_retries: int = 5
    """Maximum number of retries for API requests"""

    request_retry_base_delay: float = 1.0
    """Base delay before retrying a API request in seconds"""

    realtime_client_id: Optional[StrictStr] = None
    """Client ID for realtime service."""

    max_recursion_factor: int = 5
    """Maximum recursion factor for UDFs. This is used to limit the number of
    recursive calls to UDFs. If a UDF exceeds this limit, an error will be raised."""

    save_user_settings: StrictBool = True
    """Save per-user settings such as credentials and environment IDs."""

    default_udf_run_engine: Optional[StrictStr] = None
    """Default engine to run UDFs, one of: "local" or "remote"."""

    default_validate_imports: StrictBool = False
    """Default for whether to validate imports in UDFs before `run_local`,
    `run_batch`."""

    prompt_to_login: StrictBool = False
    """Automatically prompt the user to login when importing Fused."""

    no_login: StrictBool = False
    """If set, Fused will not attempt to login automatically when needed."""

    pyodide_async_requests: StrictBool = False
    """If set, Fused is being called inside Pyodide and should use pyodide
    for async HTTP requests."""

    cache_directory: Path = Field(default_factory=_cache_directory)
    """The base directory for storing cached results."""

    data_directory: Path = Field(default_factory=_data_directory)
    """The base directory for storing data results."""

    temp_directory: Path = Path(tempfile.gettempdir())
    """The base directory for storing temporary files."""

    never_import: StrictBool = False
    """Never import UDF code when loading UDFs."""

    gcs_secret: str = "gcs_fused"
    """Secret name for GCS credentials."""

    gcs_filename: str = "/tmp/.gcs.fused"
    """Filename for saving temporary GCS credentials to locally or in rt2 instance"""

    logging: StrictBool = Field(default=False, validate_default=True)
    """Control logging for Fused"""

    default_run_headers: Optional[Dict[str, str]] = {"X-Fused-Cache-Disable": "true"}
    """(Advanced) Default headers to include with UDF run requests."""

    fd_prefix: Optional[str] = None
    """If set, where fd:// scheme URLs will resolve to. By default will infer this from your user account."""

    @field_validator("logging")
    @classmethod
    def _validate_logging(cls, v):
        from loguru import logger

        if v:
            logger.enable("fused")
        else:
            logger.disable("fused")

        return v

    @field_serializer("cache_directory")
    def _serialize_cache_directory(self, path: Path, _info) -> str:
        return str(path)

    @field_serializer("temp_directory")
    def _serialize_temp_directory(self, path: Path, _info) -> str:
        return str(path)

    @field_validator("base_url")
    @classmethod
    def _validate_base_url(cls, v):
        reset_api()
        return v

    @field_validator("auth")
    @classmethod
    def _validate_auth(cls, v):
        from fused._auth import AUTHORIZATION

        # reset to trigger re-authentication / loading from disk to ensure setting
        # the options.auth blow has effect
        AUTHORIZATION.reset()

        return v

    @field_validator("data_directory", mode="before")
    @classmethod
    def _validate_data_directory(cls, v):
        if is_path_writable(Path(v)):
            return v
        else:
            return _data_directory()

    @property
    def base_web_url(self):
        if self.base_url == STAGING_DEFAULT_BASE_URL:
            return "https://staging.fused.io"

        parts = urlparse(self.base_url)
        return "{scheme}://{netloc}".format(**parts._asdict())

    def save(self):
        """Save Fused options to `~/.fused/settings.toml`. They will be automatically
        reloaded the next time fused-py is imported.
        """
        try:
            import rtoml

            OPTIONS_PATH.parent.mkdir(parents=True, exist_ok=True)
            # None (null) will not be serialized correctly in toml, so exclude it.
            # Any option which can be None should be None by default. Some open options
            # don't do this; should be updated to not be Optional.
            rtoml.dump(
                self.model_dump(exclude_none=True, exclude_defaults=True), OPTIONS_PATH
            )
        except ImportError:
            warnings.warn(
                FusedImportWarning("rtoml is not installed so options are not saved")
            )

    def _to_toml(self) -> str:
        try:
            import rtoml

            return rtoml.dumps(
                self.model_dump(exclude_none=True, exclude_defaults=True)
            )
        except ImportError:
            warnings.warn(
                FusedImportWarning("rtoml is not installed so options are not saved")
            )

    model_config = ConfigDict(validate_assignment=True)


def _load_options():
    if OPTIONS_PATH.exists():
        try:
            import rtoml

            return Options.model_validate(rtoml.load(OPTIONS_PATH))
        except:  # noqa E722
            warnings.warn(
                FusedIgnoredWarning(
                    f"Settings file {OPTIONS_PATH} exists but could not be loaded."
                )
            )

    if OPTIONS_JSON_PATH.exists():
        try:
            import json

            with open(OPTIONS_JSON_PATH) as json_file:
                return Options.model_validate(json.load(json_file))
        except:  # noqa E722
            warnings.warn(
                FusedIgnoredWarning(
                    f"Settings file {OPTIONS_JSON_PATH} exists but could not be loaded."
                )
            )

    return Options()


options = _load_options()
"""List global configuration options.

This object contains a set of configuration options that control global behavior of the library. This object can be used to modify the options.

Examples:
    Change the `request_timeout` option from its default value to 120 seconds:
    ```py
    fused.options.request_timeout = 120
    ```
"""


def _dev_auth():
    auth = AuthOptions()
    auth.client_id = "K5XfH4xP6PQo6weGVvPYDUiLxGu7GMdb"
    auth.client_secret = (
        "sHh339I8c8bDEPXVMRHMi9rnSrwWtawBHW1588Rm9mqGTc8F5khtvpYOha6Bxk7P"
    )
    auth.audience = "fused-python-api"
    auth.oauth_token_url = "https://fused-dev.us.auth0.com/oauth/token"
    auth.authorize_url = "https://fused-dev.us.auth0.com/authorize"
    auth.credentials_path = "~/.fused/credentials-dev"
    return auth


def env(environment_name: Union[str, int]):
    """Set the environment."""
    if isinstance(environment_name, int) or environment_name.isnumeric():
        _env = DEV_PR_DEFAULT_BASE_URL.format(environment_name)
        options.auth = _dev_auth()
        setattr(options, "base_url", _env)
        return
    else:
        # reset auth options in case we are switching from a dev env
        options.auth = AuthOptions()

    if environment_name == "dev":
        _env = DEV_DEFAULT_BASE_URL
    elif environment_name == "stg" or environment_name == "staging":
        _env = STAGING_DEFAULT_BASE_URL
    elif environment_name == "prod":
        _env = PROD_DEFAULT_BASE_URL
    elif environment_name == "unstable":
        _env = UNSTABLE_DEFAULT_BASE_URL
    else:
        raise ValueError("Available options are `dev`, `stg`, `prod`, and `unstable`.")

    # setting this will cause the global API to be reset
    setattr(options, "base_url", _env)

    # Reset realtime_client_id since it should no longer be correct for a new env.
    setattr(options, "realtime_client_id", None)



from __future__ import annotations

from pathlib import Path
from typing import Any
from zipfile import ZipFile

from fused.models.udf import AnyBaseUdf

from .._str_utils import is_uuid
from ._impl._context_impl import context_get_user_email
from ._impl._udf_ops_impl import (
    get_github_udf_from_server,
    get_step_config_from_server,
    get_step_config_from_shared_token,
    get_udf_from_code,
    get_udf_from_directory,
    get_udf_from_file,
)


def load_udf_from_fused(
    email_or_handle_or_id: str, id: str | None = None, *, cache_key: Any = None
) -> AnyBaseUdf:
    """
    Download the code of a UDF, to be run inline.

    Args:
        email_or_handle_or_id: Email or handle of the UDF's owner, or name of the UDF to import.
        id: Name of the UDF to import. If only the first argument is provided, the current user's email will be used.

    Keyword args:
        cache_key: Additional cache key for busting the UDF cache
    """
    if id is None and not is_uuid(email_or_handle_or_id):
        id = email_or_handle_or_id
        try:
            email_or_handle = context_get_user_email()
        except Exception as e:
            raise ValueError(
                "could not detect user ID from context, please specify the UDF as 'user@example.com' (or 'user'), 'udf_name'."
            ) from e
    else:
        email_or_handle = email_or_handle_or_id
    step_config = get_step_config_from_server(
        email_or_handle=email_or_handle,
        slug=id,
        cache_key=cache_key,
    )

    return step_config.udf


def load_udf_from_github(url: str, *, cache_key: Any = None) -> AnyBaseUdf:
    """
    Download the code of a UDF, to be run inline.

    Args:
        email_or_id: Email of the UDF's owner, or name of the UDF to import.
        id: Name of the UDF to import. If only the first argument is provided, the current user's email will be used.

    Keyword args:
        cache_key: Additional cache key for busting the UDF cache
    """
    step_config = get_github_udf_from_server(url=url, cache_key=cache_key)

    return step_config.udf


def load_udf_from_shared_token(token: str) -> AnyBaseUdf:
    """
    Download the code of a UDF from a shared token

    Args:
        token: the shared token for a UDF

    Raises:
        requests.HTTPError if the token is for a UDF that is not owned by the current user
    """
    return get_step_config_from_shared_token(token).udf


def load_udf_from_file(path: Path) -> AnyBaseUdf:
    """
    Load a UDF from a python file

    Args:
        path : pathlib.Path
    """
    return get_udf_from_file(path)


def load_udf_from_directory(path: Path) -> AnyBaseUdf:
    """
    Load a UDF from a python file

    Args:
        path : pathlib.Path
    """

    def _load_file(name: str) -> bytes:
        file_path = path / name
        if not file_path.exists():
            raise ValueError(
                f"Expected a file to be at {repr(file_path)}. Is this the right directory to load from?"
            )

        return file_path.read_bytes()

    return get_udf_from_directory(load_callback=_load_file)


def load_udf_from_zip(path: Path) -> AnyBaseUdf:
    """
    Load a UDF from a python file

    Args:
        path : pathlib.Path
    """
    with ZipFile(path) as zf:
        return get_udf_from_directory(load_callback=lambda f: zf.read(f))


def load_udf_from_code(code: str) -> AnyBaseUdf:
    """
    Load a UDF from raw code.

    """
    return get_udf_from_code(code)



from contextlib import contextmanager
from pathlib import Path
from typing import Any, Dict, Protocol

from loguru import logger

from fused._global_api import get_api


class ExecutionContextProtocol(Protocol):
    def __enter__(self): ...

    def __exit__(self, *exc_details): ...

    @property
    def partition_tempdir(self) -> Path:
        """A partition-level temporary directory for user use during the job.

        A new directory is provided for each file."""
        ...

    @property
    def tempdir(self) -> Path:
        """A chunk-level temporary directory for user use during the job.

        A new directory is provided for each chunk."""
        ...

    @property
    def dataset(self) -> Any:
        """The dataset (or left dataset in the case of a join)."""
        ...

    def auth_header(self, *, missing_ok: bool = False) -> Dict[str, str]:
        """Return the auth header to use for the current context."""
        ...

    @property
    def auth_token(self) -> str | None:
        """User's authentication token."""
        ...

    @property
    def auth_scheme(self) -> str | None:
        """User's authentication token scheme."""
        ...

    @property
    def user_email(self) -> str | None: ...

    @property
    def realtime_client_id(self) -> str | None: ...

    @property
    def recursion_factor(self) -> int | None: ...

    @property
    def in_realtime(self) -> bool:
        """Return True if the context is in a realtime job."""
        return False

    @property
    def in_batch(self) -> bool:
        """Return True if the context is in a batch job."""
        return False

    def get_secret(self, key: str, client_id: str | None = None):
        api = get_api()
        return api.get_secret_value(key=key, client_id=client_id)

    def list_secrets(self, client_id: str | None = None):
        api = get_api()
        return api.list_secrets(client_id=client_id)

    def set_secret(self, key: str, value: str, client_id: str | None = None):
        api = get_api()
        return api.set_secret_value(key=key, value=value, client_id=client_id)

    def delete_secret(self, key: str, client_id: str | None = None):
        api = get_api()
        return api.delete_secret_value(key=key, client_id=client_id)


GLOBAL_CONTEXT_MANAGER: ExecutionContextProtocol | None = None


def get_global_context() -> ExecutionContextProtocol | None:
    return GLOBAL_CONTEXT_MANAGER


@contextmanager
def global_context(context: ExecutionContextProtocol, *, allow_override: bool = False):
    """Set an ExecutionContext object as the global one"""
    global GLOBAL_CONTEXT_MANAGER
    if GLOBAL_CONTEXT_MANAGER is not None and not allow_override:
        logger.warning("Setting global context while it is already set")

    prev_context = GLOBAL_CONTEXT_MANAGER
    try:
        GLOBAL_CONTEXT_MANAGER = context
        yield GLOBAL_CONTEXT_MANAGER
    finally:
        GLOBAL_CONTEXT_MANAGER = prev_context


class LocalExecutionContext(ExecutionContextProtocol):
    @property
    def partition_tempdir(self) -> Path:
        raise NotImplementedError()

    @property
    def tempdir(self) -> Path:
        raise NotImplementedError()

    def auth_header(self, *, missing_ok: bool = False) -> Dict[str, str]:
        from fused._auth import AUTHORIZATION

        if AUTHORIZATION.is_configured() or not missing_ok:
            return {"Authorization": f"Bearer {AUTHORIZATION.credentials.access_token}"}
        else:
            # Not logged in and that's OK
            return {}

    @property
    def auth_token(self) -> str | None:
        return None

    @property
    def auth_scheme(self) -> str | None:
        return None

    @property
    def user_email(self) -> str | None:
        api = get_api()
        return api._whoami()["email"]

    @property
    def realtime_client_id(self) -> str | None:
        api = get_api()
        return api._automatic_realtime_client_id()

    @property
    def recursion_factor(self) -> int | None:
        return 0


context: ExecutionContextProtocol = LocalExecutionContext()
local_context = LocalExecutionContext()


GLOBAL_CONTEXT_MANAGER = local_context



# TODO: This file is no longer the most recent -- use fused.core.run_* instead
# This file is only for running non-saved (code included) UDFs
from __future__ import annotations

import json
import time
from pathlib import Path
from typing import TYPE_CHECKING, Any

import pyarrow as pa
from loguru import logger

from fused._options import options as OPTIONS
from fused._request import default_retry_strategy, session_with_retries
from fused.api.api import FusedAPI, resolve_udf_server_url
from fused.core._realtime_ops import (
    _process_response,
    _process_response_async,
    serialize_realtime_params,
)
from fused.models import AnyJobStepConfig
from fused.models.udf._eval_result import UdfEvaluationResult

from ..core._impl._realtime_ops_impl import get_recursion_factor

if TYPE_CHECKING:
    import geopandas as gpd
    import pandas as pd


def run_tile(
    x: float,
    y: float,
    z: float,
    data: None | (pd.DataFrame | gpd.GeoDataFrame | pa.Table | str | Path | Any) = None,
    step_config: AnyJobStepConfig | None = None,
    params: dict[str, str] | None = None,
    *,
    print_time: bool = False,
    client_id: str | None = None,
    cache_max_age: int | None = None,
    dtype_out_vector: str = "parquet",
    dtype_out_raster: str = "tiff",
) -> UdfEvaluationResult:  # TODO: return png
    time_start = time.perf_counter()
    # We need to get the auth headers from the FusedAPI. Don't enable set_global_api
    # to avoid messing up the user's environment.
    api = FusedAPI(set_global_api=False)

    udf_server_url = resolve_udf_server_url(client_id)

    assert step_config is not None
    # Apply parameters, if we want to step_config that's returned to have this,
    # overwrite step_config.
    step_config_with_params = step_config.set_udf(
        udf=step_config.udf, parameters=serialize_realtime_params(params)
    )

    url = f"{udf_server_url}/api/v1/run/udf/tiles/{z}/{x}/{y}"

    # Headers
    recursion_factor = get_recursion_factor()
    headers = api._generate_headers(
        {
            "Content-Type": "application/json",
            **(OPTIONS.default_run_headers or {}),
        }
    )
    headers["Fused-Recursion"] = f"{recursion_factor}"

    # Payload
    post_attr_json = {
        "data_left": data,
        "data_right": None,
        "step_config": step_config_with_params.model_dump_json(),
        "dtype_in": "json",
        # TODO remove dtype_out
        "dtype_out_vector": dtype_out_vector,
        "dtype_out_raster": dtype_out_raster,
    }

    # Params
    req_params = {
        **({"cache_max_age": int(cache_max_age)} if cache_max_age is not None else {}),
        "dtype_out_vector": dtype_out_vector,
        "dtype_out_raster": dtype_out_raster,
    }

    # Make request
    start = time.time()

    with session_with_retries() as session:
        r = session.post(
            url=url,
            params=req_params,
            json=post_attr_json,
            headers=headers,
            timeout=OPTIONS.request_timeout,
        )

    end = time.time()
    if print_time:
        logger.info(f"Time in request: {end - start}")

    time_end = time.perf_counter()
    time_taken_seconds = time_end - time_start

    return _process_response(
        r,
        step_config=step_config_with_params,
        time_taken_seconds=time_taken_seconds,
    )


def run(
    step_config: AnyJobStepConfig | None = None,
    params: dict[str, str] | None = None,
    *,
    print_time: bool = False,
    read_options: dict | None = None,
    client_id: str | None = None,
    cache_max_age: int | None = None,
    dtype_out_vector: str = "parquet",
    dtype_out_raster: str = "tiff",
) -> pd.DataFrame:
    """Run a UDF.

    Args:
        step_config: AnyJobStepConfig.
        params: Additional parameters to pass to the UDF. Must be JSON serializable.

    Keyword Args:
        print_time: If True, print the amount of time taken in the request.
        read_options: If not None, options for reading `df` that will be passed to GeoPandas.
    """
    # TODO: This function is too complicated

    time_start = time.perf_counter()
    # We need to get the auth headers from the FusedAPI. Don't enable set_global_api
    # to avoid messing up the user's environment.
    api = FusedAPI(set_global_api=False)

    udf_server_url = resolve_udf_server_url(client_id)

    assert step_config is not None
    # Apply parameters, if we want to step_config that's returned to have this,
    # overwrite step_config.
    step_config_with_params = step_config.set_udf(
        udf=step_config.udf, parameters=serialize_realtime_params(params)
    )

    # Note: Custom UDF uses the json POST attribute.
    url = f"{udf_server_url}/api/v1/run/udf"

    # This is the body for when step_config_with_params.type == "udf".
    body = {
        "data_left": None,
        "step_config": step_config_with_params.model_dump_json(),
        "dtype_in": "json",
        # TODO remove dtype_out
        "dtype_out_vector": dtype_out_vector,
        "dtype_out_raster": dtype_out_raster,
    }

    method = "POST"
    post_attr_json = body

    recursion_factor = get_recursion_factor()
    post_attr_headers = api._generate_headers(
        {
            "Content-Type": "application/json",
            **(OPTIONS.default_run_headers or {}),
        }
    )
    post_attr_headers["Fused-Recursion"] = f"{recursion_factor}"

    req_params = {
        **({"cache_max_age": int(cache_max_age)} if cache_max_age is not None else {}),
        "dtype_out_vector": dtype_out_vector,
        "dtype_out_raster": dtype_out_raster,
    }

    # Make request
    start = time.time()

    with session_with_retries() as session:
        r = session.request(
            method=method,
            url=url,
            params=req_params,
            json=post_attr_json,
            headers=post_attr_headers,
            timeout=OPTIONS.request_timeout,
        )

    end = time.time()
    if print_time:
        logger.info(f"Time in request: {end - start}")

    time_end = time.perf_counter()
    time_taken_seconds = time_end - time_start

    return _process_response(
        r, step_config=step_config_with_params, time_taken_seconds=time_taken_seconds
    )


async def run_tile_async(
    x: float,
    y: float,
    z: float,
    step_config: AnyJobStepConfig | None = None,
    params: dict[str, str] | None = None,
    *,
    print_time: bool = False,
    client_id: str | None = None,
    cache_max_age: int | None = None,
    dtype_out_vector: str = "parquet",
    dtype_out_raster: str = "tiff",
) -> UdfEvaluationResult:  # TODO: return png
    time_start = time.perf_counter()
    # We need to get the auth headers from the FusedAPI. Don't enable set_global_api
    # to avoid messing up the user's environment.
    api = FusedAPI(set_global_api=False)

    udf_server_url = resolve_udf_server_url(client_id)

    assert step_config is not None
    # Apply parameters, if we want to step_config that's returned to have this,
    # overwrite step_config.
    step_config_with_params = step_config.set_udf(
        udf=step_config.udf, parameters=serialize_realtime_params(params)
    )

    url = f"{udf_server_url}/api/v1/run/udf/tiles/{z}/{x}/{y}"

    # Headers
    recursion_factor = get_recursion_factor()
    headers = api._generate_headers(
        {
            "Content-Type": "application/json",
            **(OPTIONS.default_run_headers or {}),
        }
    )
    headers["Fused-Recursion"] = f"{recursion_factor}"

    # Payload
    post_attr_json = {
        "data_left": None,
        "data_right": None,
        "step_config": step_config_with_params.model_dump_json(),
        "dtype_in": "json",
        # TODO remove dtype_out
        "dtype_out_vector": dtype_out_vector,
        "dtype_out_raster": dtype_out_raster,
    }

    # Params
    req_params = {
        **({"cache_max_age": int(cache_max_age)} if cache_max_age is not None else {}),
        "dtype_out_vector": dtype_out_vector,
        "dtype_out_raster": dtype_out_raster,
    }

    # Make request
    start = time.time()

    if OPTIONS.pyodide_async_requests:
        import pyodide.http
        import yarl

        url_with_params = yarl.URL(url, encoded=True).with_query(req_params)
        r = await pyodide.http.pyfetch(
            str(url_with_params),
            method="POST",
            headers=headers,
            body=json.dumps(post_attr_json),
            # TODO: timeout
        )
    else:
        import aiohttp

        async with aiohttp.ClientSession() as session:
            r = await session.post(
                url=url,
                params=req_params,
                json=post_attr_json,
                headers=headers,
                retry=default_retry_strategy,
                # TODO: timeout
            )

    end = time.time()
    if print_time:
        logger.info(f"Time in request: {end - start}")

    time_end = time.perf_counter()
    time_taken_seconds = time_end - time_start

    return await _process_response_async(
        r, step_config=step_config_with_params, time_taken_seconds=time_taken_seconds
    )


async def run_async(
    step_config: AnyJobStepConfig | None = None,
    params: dict[str, str] | None = None,
    *,
    print_time: bool = False,
    client_id: str | None = None,
    cache_max_age: int | None = None,
    dtype_out_vector: str = "parquet",
    dtype_out_raster: str = "tiff",
) -> pd.DataFrame:
    """Run a UDF over a DataFrame.

    Args:
        step_config: AnyJobStepConfig.
        params: Additional parameters to pass to the UDF. Must be JSON serializable.

    Keyword Args:
        print_time: If True, print the amount of time taken in the request.
    """
    # TODO: This function is too complicated

    time_start = time.perf_counter()
    # We need to get the auth headers from the FusedAPI. Don't enable set_global_api
    # to avoid messing up the user's environment.
    api = FusedAPI(set_global_api=False)

    udf_server_url = resolve_udf_server_url(client_id)

    assert step_config is not None
    # Apply parameters, if we want to step_config that's returned to have this,
    # overwrite step_config.
    step_config_with_params = step_config.set_udf(
        udf=step_config.udf, parameters=serialize_realtime_params(params)
    )

    # Note: Custom UDF uses the json POST attribute.
    url = f"{udf_server_url}/api/v1/run/udf"

    # This is the body for when step_config_with_params.type == "udf".
    body = {
        "data_left": None,
        "step_config": step_config_with_params.model_dump_json(),
        "dtype_in": "json",
        # TODO remove dtype_out
        "dtype_out_vector": dtype_out_vector,
        "dtype_out_raster": dtype_out_raster,
    }

    post_attr_json = body

    recursion_factor = get_recursion_factor()
    post_attr_headers = api._generate_headers(
        {
            "Content-Type": "application/json",
            **(OPTIONS.default_run_headers or {}),
        }
    )
    post_attr_headers["Fused-Recursion"] = f"{recursion_factor}"

    req_params = {
        **({"cache_max_age": int(cache_max_age)} if cache_max_age is not None else {}),
        "dtype_out_vector": dtype_out_vector,
        "dtype_out_raster": dtype_out_raster,
    }
    # Make request
    start = time.time()

    if OPTIONS.pyodide_async_requests:
        import pyodide.http
        import yarl

        url_with_params = yarl.URL(url, encoded=True).with_query(req_params)
        r = await pyodide.http.pyfetch(
            str(url_with_params),
            method="POST",
            headers=post_attr_headers,
            body=json.dumps(post_attr_json),
            # TODO: timeout
        )
    else:
        import aiohttp

        async with aiohttp.ClientSession() as session:
            r = await session.post(
                url=url,
                params=req_params,
                json=post_attr_json,
                headers=post_attr_headers,
                retry=default_retry_strategy,
                # TODO: timeout
            )

    end = time.time()
    if print_time:
        logger.info(f"Time in request: {end - start}")

    time_end = time.perf_counter()
    time_taken_seconds = time_end - time_start

    return await _process_response_async(
        r, step_config=step_config_with_params, time_taken_seconds=time_taken_seconds
    )
